from pydantic.dataclasses import dataclass

from ..base.evaluator_config import EvaluatorConfig


@dataclass
class VisualizationConfig(EvaluatorConfig):

    """

    Config class for the visualization module.

    Args:
        batch_size (int) : The batch size to use in the evaluation. Default to 512
        wandb_path (str) : The user can provide the path of the wandb run with a
            format 'entity/projet_name/run_id' where the metrics should be logged.
            For an existing run (the training run), the info can be found in the training dir
            (in wandb_info.json)
            at the end of training (if wandb was used) or on the hugging_face webpage of the run.
            Otherwise the user can create a new wandb run and get the path with :
                `
                import wandb
                run = wandb.init(entity = your_entity, project=your_project)
                wandb_path = run.path

            If None are provided, the metrics are not logged on wandb.
            Default to None.
        n_samples(int) : The number of samples to generate per modality and per data_point for
            conditional generation. Default to 5.
        n_data_cond (int) : The number of datapoints to use for conditional generation. Default to 5
    """

    n_samples: int = 5
    n_data_cond: int = 5



import datetime
import json
import logging
import os
from copy import deepcopy
from typing import List, Optional

import torch
import torch.distributed as dist
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
from PIL import Image
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from torchvision.utils import make_grid

from ...data import MultimodalBaseDataset
from ...data.datasets.utils import adapt_shape
from ...data.utils import set_inputs_to_device
from ...models import BaseMultiVAE
from .base_trainer_config import BaseTrainerConfig
from .callbacks import (
    CallbackHandler,
    MetricConsolePrinterCallback,
    ProgressBarCallback,
    TrainingCallback,
)
from .utils import set_seed, update_dict

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class BaseTrainer:
    """Base class to perform model training.

    Args:
        model (BaseMultiVAE): A instance of :class:`~multivae.models.BaseMultiVAE` to train.

        train_dataset (MultimodalBaseDataset): The training dataset of type
            :class:`~multivae.data.datasets.MultimodalBaseDataset`

        eval_dataset (MultimodalBaseDataset): The evaluation dataset of type
            :class:`~multivae.data.datasets.MultimodalBaseDataset`

        training_config (BaseTrainerConfig): The training arguments summarizing the main
            parameters used for training. If None, a basic training instance of
            :class:`BaseTrainerConfig` is used. Default: None.

        callbacks (List[~pythae.trainers.training_callbacks.TrainingCallback]):
            A list of callbacks to use during training.

        checkpoint (str) : The directory path of the checkpoint generated by the `save_checkpoint`
            function. Default to None.
    """

    def __init__(
        self,
        model: BaseMultiVAE,
        train_dataset: MultimodalBaseDataset,
        eval_dataset: Optional[MultimodalBaseDataset] = None,
        training_config: Optional[BaseTrainerConfig] = None,
        callbacks: List[TrainingCallback] = None,
        checkpoint: str = None,
    ):
        if training_config is None:
            if checkpoint is None:
                training_config = BaseTrainerConfig()
            else:
                training_config = BaseTrainerConfig.from_json_file(
                    os.path.join(checkpoint, "training_config.json")
                )

        if training_config.output_dir is None:
            output_dir = "dummy_output_dir"
            training_config.output_dir = output_dir

        self.training_config = training_config
        self.model_config = model.model_config
        self.model_name = model.model_name

        # for distributed training
        self.world_size = self.training_config.world_size
        self.local_rank = self.training_config.local_rank
        self.rank = self.training_config.rank
        self.dist_backend = self.training_config.dist_backend

        if self.world_size > 1:
            self.distributed = True
        else:
            self.distributed = False

        if self.distributed:
            device = self._setup_devices()

        else:
            device = (
                "cuda"
                if torch.cuda.is_available() and not self.training_config.no_cuda
                else "cpu"
            )

        self.device = device

        if checkpoint is not None:
            model = model.load_from_folder(checkpoint)

        # place model on device
        model = model.to(device)
        model.device = device

        if self.distributed:
            model = DDP(model, device_ids=[self.local_rank])

        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset

        # Define the loaders
        train_loader = self.get_train_dataloader(train_dataset)

        if eval_dataset is not None:
            eval_loader = self.get_eval_dataloader(eval_dataset)

        else:
            logger.info(
                "! No eval dataset provided ! -> keeping best model on train.\n"
            )
            self.training_config.keep_best_on_train = True
            eval_loader = None

        self.train_loader = train_loader
        self.eval_loader = eval_loader
        self.callbacks = callbacks

        # run sanity check on the model
        self._run_model_sanity_check(model, train_loader)

        if self.is_main_process:
            logger.info("Model passed sanity check !\n" "Ready for training.\n")

        self.model = model

        if checkpoint is None:
            self.prepare_training()
        else:
            self.resume_training(checkpoint)

    @property
    def is_main_process(self):
        if self.rank == 0 or self.rank == -1:
            return True
        else:
            return False

    def _setup_devices(self):
        """Sets up the devices to perform distributed training."""

        if dist.is_available() and dist.is_initialized() and self.local_rank == -1:
            logger.warning(
                "torch.distributed process group is initialized, but local_rank == -1. "
            )
        if self.training_config.no_cuda:
            self._n_gpus = 0
            device = "cpu"

        else:
            torch.cuda.set_device(self.local_rank)
            device = torch.device("cuda", self.local_rank)

            if not dist.is_initialized():
                dist.init_process_group(
                    backend=self.dist_backend,
                    init_method="env://",
                    world_size=self.world_size,
                    rank=self.rank,
                )

        return device

    def get_train_dataloader(
        self, train_dataset: MultimodalBaseDataset
    ) -> torch.utils.data.DataLoader:
        if self.distributed:
            train_sampler = DistributedSampler(
                train_dataset, num_replicas=self.world_size, rank=self.rank
            )
        else:
            train_sampler = None
        return DataLoader(
            dataset=train_dataset,
            batch_size=self.training_config.per_device_train_batch_size,
            num_workers=self.training_config.train_dataloader_num_workers,
            shuffle=True,
            sampler=train_sampler,
            drop_last=self.training_config.drop_last,
        )

    def get_eval_dataloader(
        self, eval_dataset: MultimodalBaseDataset
    ) -> torch.utils.data.DataLoader:
        if self.distributed:
            eval_sampler = DistributedSampler(
                eval_dataset, num_replicas=self.world_size, rank=self.rank
            )
        else:
            eval_sampler = None
        return DataLoader(
            dataset=eval_dataset,
            batch_size=self.training_config.per_device_eval_batch_size,
            num_workers=self.training_config.eval_dataloader_num_workers,
            shuffle=(eval_sampler is None),
            sampler=eval_sampler,
        )

    def set_optimizer(self):
        optimizer_cls = getattr(optim, self.training_config.optimizer_cls)
        logger.info(
            f"Setting the optimizer with learning rate {self.training_config.learning_rate}"
        )
        if self.training_config.optimizer_params is not None:
            optimizer = optimizer_cls(
                self.model.parameters(),
                lr=self.training_config.learning_rate,
                **self.training_config.optimizer_params,
            )
        else:
            optimizer = optimizer_cls(
                self.model.parameters(), lr=self.training_config.learning_rate
            )

        self.optimizer = optimizer

    def set_scheduler(self):
        if self.training_config.scheduler_cls is not None:
            scheduler_cls = getattr(lr_scheduler, self.training_config.scheduler_cls)

            if self.training_config.scheduler_params is not None:
                scheduler = scheduler_cls(
                    self.optimizer, **self.training_config.scheduler_params
                )
            else:
                scheduler = scheduler_cls(self.optimizer)

        else:
            scheduler = None

        self.scheduler = scheduler

    def _set_output_dir(self):
        # Create folder
        if not os.path.exists(self.training_config.output_dir) and self.is_main_process:
            os.makedirs(self.training_config.output_dir, exist_ok=True)
            logger.info(
                f"Created {self.training_config.output_dir} folder since did not exist.\n"
            )

        self._training_signature = (
            str(datetime.datetime.now())[0:19].replace(" ", "_").replace(":", "-")
        )

        training_dir = os.path.join(
            self.training_config.output_dir,
            f"{self.model_name}_training_{self._training_signature}",
        )

        self.training_dir = training_dir

        if not os.path.exists(training_dir) and self.is_main_process:
            os.makedirs(training_dir, exist_ok=True)
            logger.info(
                f"Created {training_dir}. \n"
                "Training config, checkpoints and final model will be saved here.\n"
            )

    def _get_file_logger(self, log_output_dir):
        log_dir = log_output_dir

        # if dir does not exist create it
        if not os.path.exists(log_dir) and self.is_main_process:
            os.makedirs(log_dir, exist_ok=True)
            logger.info(f"Created {log_dir} folder since did not exists.")
            logger.info("Training logs will be recodered here.\n")
            logger.info(" -> Training can be monitored here.\n")

        # create and set logger
        log_name = f"training_logs_{self._training_signature}"

        file_logger = logging.getLogger(log_name)
        file_logger.setLevel(logging.INFO)
        f_handler = logging.FileHandler(
            os.path.join(log_dir, f"training_logs_{self._training_signature}.log")
        )
        f_handler.setLevel(logging.INFO)
        file_logger.addHandler(f_handler)

        # Do not output logs in the console
        file_logger.propagate = False

        return file_logger

    def _setup_callbacks(self):
        if self.callbacks is None:
            self.callbacks = [TrainingCallback()]

        self.callback_handler = CallbackHandler(
            callbacks=self.callbacks, model=self.model
        )

        self.callback_handler.add_callback(ProgressBarCallback())
        self.callback_handler.add_callback(MetricConsolePrinterCallback())

    def _run_model_sanity_check(self, model, loader):
        try:
            inputs = next(iter(loader))
            train_dataset = set_inputs_to_device(inputs, device=self.device)
            model(train_dataset)

        except Exception as e:
            raise Exception(
                "Error when calling forward method from model. Potential issues: \n"
                " - Wrong model architecture -> check encoder, decoder and metric architecture if "
                "you provide yours \n"
                " - The data input dimension provided is wrong -> when no encoder, decoder or metric "
                "provided, a network is built automatically but requires the shape of the flatten "
                "input data.\n"
                f"Exception raised: {type(e)} with message: " + str(e)
            ) from e

    def _optimizers_step(self, model_output=None):
        loss = model_output.loss

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

    def _schedulers_step(self, metrics=None):
        if self.scheduler is None:
            pass

        elif isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau):
            self.scheduler.step(metrics)

        else:
            self.scheduler.step()

    def prepare_training(self):
        """Sets up the trainer for training"""
        # set random seed
        set_seed(self.training_config.seed)

        # set optimizer
        self.set_optimizer()

        # set scheduler
        self.set_scheduler()

        # create folder for saving
        self._set_output_dir()

        # set callbacks
        self._setup_callbacks()

        # set already_trained epochs
        self.trained_epochs = 0
        self.best_train_loss = torch.inf
        self.best_eval_loss = torch.inf
        # set up the best_model
        self._best_model = deepcopy(self.model)

    def resume_training(self, checkpoint):
        """Sets up the trainer for training"""
        with open(os.path.join(checkpoint, "info_checkpoint.json"), "r") as fp:
            dict_checkpoint = json.load(fp)

        # set random seed
        set_seed(self.training_config.seed)

        # set optimizer
        self.set_optimizer()
        self.optimizer.load_state_dict(
            torch.load(
                os.path.join(checkpoint, "optimizer.pt"), map_location=self.device
            )
        )

        # set scheduler
        self.set_scheduler()
        if self.scheduler is not None:
            self.scheduler.load_state_dict(
                torch.load(
                    os.path.join(checkpoint, "scheduler.pt"), map_location=self.device
                )
            )

        # create folder for saving
        self.training_dir = dict_checkpoint["training_dir"]

        # set callbacks
        self._setup_callbacks()

        # set already_trained epochs
        self.trained_epochs = dict_checkpoint["trained_epochs"]
        self.best_train_loss = dict_checkpoint["best_train_loss"]
        self.best_eval_loss = dict_checkpoint["best_eval_loss"]

        # set up the best_model
        self._best_model = deepcopy(self.model)

    def prepare_train_step(self, epoch, best_train_loss, best_eval_loss):
        """
        Function to operate changes between train_steps such as resetting the optimizer and
        the best losses values.
        """
        return best_train_loss, best_eval_loss

    def train(self, log_output_dir: str = None):
        """This function is the main training function

        Args:
            log_output_dir (str): The path in which the log will be stored
            start_epoch (int) : The first epoch to do. Is useful in case of restarting a
                training after saving a checkpoint. Default to 1.
        """

        self.callback_handler.on_train_begin(
            training_config=self.training_config, model_config=self.model_config
        )

        log_verbose = False

        msg = (
            f"Training params:\n - max_epochs: {self.training_config.num_epochs}\n"
            " - per_device_train_batch_size: "
            f"{self.training_config.per_device_train_batch_size}\n"
            " - per_device_eval_batch_size: "
            f"{self.training_config.per_device_eval_batch_size}\n"
            f" - checkpoint saving every: {self.training_config.steps_saving}\n"
            f"Optimizer: {self.optimizer}\n"
            f"Scheduler: {self.scheduler}\n"
        )

        if self.is_main_process:
            logger.info(msg)

        # set up log file
        if log_output_dir is not None and self.is_main_process:
            log_verbose = True
            file_logger = self._get_file_logger(log_output_dir=log_output_dir)

            file_logger.info(msg)

        if self.is_main_process:
            logger.info("Successfully launched training !\n")

        for epoch in range(
            self.trained_epochs + 1, self.training_config.num_epochs + 1
        ):
            self.callback_handler.on_epoch_begin(
                training_config=self.training_config,
                epoch=epoch,
                train_loader=self.train_loader,
                eval_loader=self.eval_loader,
            )
            self.best_train_loss, self.best_eval_loss = self.prepare_train_step(
                epoch, self.best_train_loss, self.best_eval_loss
            )
            epoch_train_loss, epoch_metrics = self.train_step(epoch)
            metrics = {"train_" + k: epoch_metrics[k] for k in epoch_metrics}
            metrics["train_epoch_loss"] = epoch_train_loss

            if self.eval_dataset is not None:
                epoch_eval_loss, epoch_eval_metrics = self.eval_step(epoch)
                metrics["eval_epoch_loss"] = epoch_eval_loss
                update_dict(
                    metrics,
                    {"eval_" + k: epoch_eval_metrics[k] for k in epoch_eval_metrics},
                )
                self._schedulers_step(epoch_eval_loss)

            else:
                epoch_eval_loss = self.best_eval_loss
                self._schedulers_step(epoch_train_loss)
            if epoch <= self.training_config.start_keep_best_epoch:
                # save the model, don't keep track of the best loss
                best_model = deepcopy(self.model)
                self._best_model = best_model
                logger.info("New best model saved!")

            elif (
                epoch_eval_loss < self.best_eval_loss
                and not self.training_config.keep_best_on_train
            ):
                self.best_eval_loss = epoch_eval_loss
                best_model = deepcopy(self.model)
                self._best_model = best_model
                logger.info("New best model saved!")

            elif (
                epoch_train_loss < self.best_train_loss
                and self.training_config.keep_best_on_train
            ):
                self.best_train_loss = epoch_train_loss
                best_model = deepcopy(self.model)
                self._best_model = best_model

            if (
                self.training_config.steps_predict is not None
                and (epoch % self.training_config.steps_predict == 0 or epoch == 1)
                and self.is_main_process
            ):
                reconstructions = self.predict(self._best_model, epoch)

                self.callback_handler.on_prediction_step(
                    self.training_config,
                    reconstructions=reconstructions,
                    global_step=epoch,
                )
            self.callback_handler.on_epoch_end(training_config=self.training_config)

            # save checkpoints
            if (
                self.training_config.steps_saving is not None
                and epoch % self.training_config.steps_saving == 0
            ):
                if self.is_main_process:
                    self.save_checkpoint(
                        model=self._best_model, dir_path=self.training_dir, epoch=epoch
                    )
                    logger.info(f"Saved checkpoint at epoch {epoch}\n")

                    if log_verbose:
                        file_logger.info(f"Saved checkpoint at epoch {epoch}\n")

            self.callback_handler.on_log(
                self.training_config,
                metrics,
                logger=logger,
                global_step=epoch,
                rank=self.rank,
            )

        final_dir = os.path.join(self.training_dir, "final_model")

        if self.is_main_process:
            self.save_model(self._best_model, dir_path=final_dir)

            logger.info("Training ended!")
            logger.info(f"Saved final model in {final_dir}")

        if self.distributed:
            dist.destroy_process_group()

        self.callback_handler.on_train_end(self.training_config)

    def eval_step(self, epoch: int):
        """Perform an evaluation step

        Parameters:
            epoch (int): The current epoch number

        Returns:
            (torch.Tensor): The evaluation loss
        """

        self.callback_handler.on_eval_step_begin(
            training_config=self.training_config,
            eval_loader=self.eval_loader,
            epoch=epoch,
            rank=self.rank,
        )

        self.model.eval()

        epoch_loss = 0
        epoch_metrics = {}

        for inputs in self.eval_loader:
            inputs = set_inputs_to_device(inputs, device=self.device)

            try:
                with torch.no_grad():
                    model_output = self.model(
                        inputs,
                        epoch=epoch,
                        dataset_size=len(self.eval_loader.dataset),
                        uses_ddp=self.distributed,
                    )

            except RuntimeError:
                model_output = self.model(
                    inputs,
                    epoch=epoch,
                    dataset_size=len(self.eval_loader.dataset),
                    uses_ddp=self.distributed,
                )

            loss = model_output.loss

            epoch_loss += loss.item()
            update_dict(epoch_metrics, model_output.metrics)

            if epoch_loss != epoch_loss:
                raise ArithmeticError("NaN detected in eval loss")

            self.callback_handler.on_eval_step_end(training_config=self.training_config)

        epoch_loss /= len(self.eval_loader)
        epoch_metrics = {
            k: epoch_metrics[k] / len(self.eval_loader) for k in epoch_metrics
        }
        return epoch_loss, epoch_metrics

    def train_step(self, epoch: int):
        """The trainer performs training loop over the train_loader.

        Parameters:
            epoch (int): The current epoch number

        Returns:
            (torch.Tensor): The step training loss
        """
        self.callback_handler.on_train_step_begin(
            training_config=self.training_config,
            train_loader=self.train_loader,
            epoch=epoch,
            rank=self.rank,
        )

        # set model in train model
        self.model.train()

        epoch_loss = 0
        epoch_model_metrics = {}
        batch_idx = 0
        for inputs in self.train_loader:
            inputs = set_inputs_to_device(inputs, device=self.device)

            model_output = self.model(
                inputs,
                epoch=epoch,
                dataset_size=len(self.train_loader.dataset),
                uses_ddp=self.distributed,
                batch_ratio=(batch_idx) / len(self.train_loader),
            )

            self._optimizers_step(model_output)

            loss = model_output.loss
            epoch_loss += loss.item()
            update_dict(epoch_model_metrics, model_output.metrics)

            if epoch_loss != epoch_loss:
                raise ArithmeticError("NaN detected in train loss")

            self.callback_handler.on_train_step_end(
                training_config=self.training_config
            )
            batch_idx += 1
        # Allows model updates if needed
        if self.distributed:
            self.model.module.update()
        else:
            self.model.update()

        epoch_loss /= len(self.train_loader)
        epoch_model_metrics = {
            k: epoch_model_metrics[k] / len(self.train_loader)
            for k in epoch_model_metrics
        }

        return epoch_loss, epoch_model_metrics

    def save_model(self, model: BaseMultiVAE, dir_path: str):
        """This method saves the final model along with the config files

        Args:
            model (BaseMultiVAE): The model to be saved
            dir_path (str): The folder where the model and config files should be saved
        """

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        # save model
        if self.distributed:
            model.module.save(dir_path)

        else:
            model.save(dir_path)

        # save training config
        self.training_config.save_json(dir_path, "training_config")

        self.callback_handler.on_save(self.training_config, dir_path=dir_path)

    def save_checkpoint(self, model: BaseMultiVAE, dir_path, epoch: int):
        """Saves a checkpoint alowing to restart training from here

        Args:
            dir_path (str): The folder where the checkpoint should be saved
            epochs_signature (int): The epoch number"""

        checkpoint_dir = os.path.join(dir_path, f"checkpoint_epoch_{epoch}")

        if not os.path.exists(checkpoint_dir):
            os.makedirs(checkpoint_dir)

        # save optimizer
        torch.save(
            deepcopy(self.optimizer.state_dict()),
            os.path.join(checkpoint_dir, "optimizer.pt"),
        )

        # save scheduler
        if self.scheduler is not None:
            torch.save(
                deepcopy(self.scheduler.state_dict()),
                os.path.join(checkpoint_dir, "scheduler.pt"),
            )

        # save model
        if self.distributed:
            model.module.save(checkpoint_dir)

        else:
            model.save(checkpoint_dir)

        # save training config
        self.training_config.save_json(checkpoint_dir, "training_config")

        # save info about checkpoint
        info = dict(
            training_dir=self.training_dir,
            trained_epochs=epoch,
            best_train_loss=self.best_train_loss,
            best_eval_loss=self.best_eval_loss,
        )
        with open(os.path.join(checkpoint_dir, "info_checkpoint.json"), "w+") as fp:
            json.dump(info, fp, sort_keys=True, indent=4)

        self.callback_handler.on_save_checkpoint(
            self.training_config, checkpoint_dir=checkpoint_dir
        )

    def predict(self, model: BaseMultiVAE, epoch: int, n_data=8):
        model.eval()

        if self.eval_dataset is not None:
            inputs = next(iter(DataLoader(self.eval_dataset, batch_size=n_data)))
            inputs = set_inputs_to_device(inputs, self.device)
        else:
            inputs = next(iter(DataLoader(self.train_dataset, batch_size=n_data)))
            inputs = set_inputs_to_device(inputs, self.device)

        # recon_dir = self.training_dir + '/reconstructions/'
        # os.makedirs(recon_dir,exist_ok=True)
        all_recons = dict()
        for mod in inputs.data:
            recon = model.predict(
                inputs, mod, "all", N=8, flatten=True, ignore_incomplete=True
            )
            recon["true_data"] = inputs.data[mod]
            recon, shape = adapt_shape(recon)
            recon_image = [recon["true_data"]] + [
                recon[m] for m in recon if m != "true_data"
            ]
            recon_image = torch.cat(recon_image)

            # Transform to PIL format
            recon_image = make_grid(recon_image, nrow=n_data)
            # Add 0.5 after unnormalizing to [0, 255] to round to nearest integer
            ndarr = (
                recon_image.mul(255)
                .add_(0.5)
                .clamp_(0, 255)
                .permute(1, 2, 0)
                .to("cpu", torch.uint8)
                .numpy()
            )
            recon_image = Image.fromarray(ndarr)

            all_recons[mod] = recon_image
        return all_recons



"""Base sampler config, adapted from Pythae's Base sampler config for multimodal data. """

from pydantic.dataclasses import dataclass
from pythae.config import BaseConfig


@dataclass
class BaseSamplerConfig(BaseConfig):
    """
    BaseSampler config class.
    """

    pass



import logging
import math
import os
import tempfile
from typing import Literal

import numpy as np
import torch
from PIL import Image
from pythae.data.datasets import DatasetOutput
from torchvision.datasets.utils import download_and_extract_archive
from torchvision.transforms import transforms

from .base import MultimodalBaseDataset

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)

class MMNISTDataset(MultimodalBaseDataset):
    """
    Multimodal MMNIST Dataset to load the Polymnist Dataset from
    'Generalized Multimodal Elbo' Sutter et al 2021.

    """

    def __init__(
        self,
        data_path : str,
        transform  = None,
        target_transform = None,
        split : Literal['train', 'test'] ="train",
        download :bool =False,
        missing_ratio : float = 0,
        keep_incomplete : bool = True
    ):
        """
        Args: 
            data_path (str) : The path where to find the MMNIST folder containing the folders 'train' or 'test'.
                The data used is the one that can be downloaded from https://zenodo.org/record/4899160#.YLn0rKgzaHu
                If data_path doesn't contain the dataset and download is set to True, then the data can be downloaded 
                for you using gdown.
            transform: tranforms on colored MNIST digits.
            target_transform: transforms on labels.
            split (Literal['train', 'test']). Which part of the data to use.
            download (bool). Autorization to download the data if it is missing at the specified location.
            missing_ratio (float between 0 and 1) : To create an partially observed dataset, specify a missing ratio > 0 and <= 1. 
                Default to 0  : No missing data. 
            keep_incomplete (bool) : For a partially observed dataset, there are two options. 
                Either keep all the samples and masks to train with incomplete data (set keep_incomplete to True)
                or only keep complete samples (keep_incomplete = False). 
                Default to True.

        """

        if isinstance(data_path, str):
            data_path = os.path.expanduser(data_path)
            

        unimodal_datapaths = [os.path.join(data_path , "MMNIST" , split , f"m{i}.pt") for i in range(5)]
        self.num_modalities = len(unimodal_datapaths)
        self.unimodal_datapaths = unimodal_datapaths
        self.transform = transform
        self.target_transform = target_transform
        self.download = download
        self.missing_ratio=missing_ratio
        self.keep_incomplete = keep_incomplete

        self.__check_or_download_data__(data_path, unimodal_datapaths)

        self.m0 = torch.load(unimodal_datapaths[0])
        self.m1 = torch.load(unimodal_datapaths[1])
        self.m2 = torch.load(unimodal_datapaths[2])
        self.m3 = torch.load(unimodal_datapaths[3])
        self.m4 = torch.load(unimodal_datapaths[4])
        
        self.images_dict = {
            "m0": self.m0,
            "m1": self.m1,
            "m2": self.m2,
            "m3": self.m3,
            "m4": self.m4,
        }

        label_datapaths = os.path.join(data_path, 'MMNIST',split, 'labels.pt')

        self.labels = torch.load(label_datapaths)

        assert self.m0.shape[0] == self.labels.shape[0]
        self.num_files = self.labels.shape[0]
        
        if missing_ratio > 0 and self.keep_incomplete:
            self.masks = {}
            for i in range(5):
                # randomly define the missing samples. 
                self.masks[f'm{i}'] = torch.bernoulli(torch.ones((self.num_files,))*(1-missing_ratio),
                                                      generator=torch.Generator().manual_seed(i)).bool()
                
            self.masks['m0']=torch.ones((self.num_files,)) # ensure there is at least one modality
                                                           # available for all samples
            
            # To be sure, also erase the content of the masked samples
            for k in self.masks:
                reverse_dim_order = tuple(np.arange(len(self.images_dict[k].shape))[::-1])
                self.images_dict[k] = self.images_dict[k].permute(*reverse_dim_order)
                # now the batch dimension is last
                self.images_dict[k] *= self.masks[k].float() # erase missing samples
                # put dimensions back in order
                self.images_dict[k] = self.images_dict[k].permute(*reverse_dim_order)
                
            
            
    def __check_or_download_data__(self, data_path, unimodal_datapaths):
        # TODO : test this function
        if not os.path.exists(unimodal_datapaths[0]) and self.download:
            tempdir = tempfile.mkdtemp()
            logger.info(f'Downloading the PolyMNIST dataset into {data_path}'
                        ' Along with the dataset, the classifiers and inception networks are also downloaded.')
            download_and_extract_archive(
                url='https://zenodo.org/record/4899160/files/PolyMNIST.zip',
                download_root=tempdir,extract_root=data_path)
            
        elif not os.path.exists(unimodal_datapaths[0]) and not self.download:
            raise AttributeError(
                "The PolyMNIST dataset is not available at the"
                " given datapath and download is set to False."
                "Set download to True or place the dataset"
                " in the data_path folder."
            )

    def __getitem__(self, index):
        """
        Returns a tuple (images, labels) where each element is a list of
        length `self.num_modalities`.
        """
        images_dict = { k: self.images_dict[k][index] for k in self.images_dict
        }
        if self.missing_ratio == 0 or not self.keep_incomplete:
            return DatasetOutput(data=images_dict, labels=self.labels[index])
        else :
            masks_dict = {k : self.masks[k][index] for k in self.masks}
                        
            return DatasetOutput(
                data = images_dict,
                labels = self.labels[index],
                masks = masks_dict
            )   

        
    def __len__(self):
        
        if self.missing_ratio == 0 or self.keep_incomplete:
            return self.num_files
        else :
            # Reduce the lenght using the proportion of complete samples
            # that corresponds to missing_ratio
            new_length = math.ceil((1 - self.missing_ratio)**4 * self.num_files)
            return new_length







import logging
from typing import List, Optional

import torch
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.models.jnf_dcca import JNFDcca
from multivae.trainers.base.callbacks import TrainingCallback

from ..base import BaseTrainer
from .add_dcca_trainer_config import AddDccaTrainerConfig

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class AddDccaTrainer(BaseTrainer):
    """A specific trainer that handles the training of the DCCA module
    that is part of the JNFDcca model.

    Args:
        model (BaseMultiVAE): A instance of :class:`~multivae.models.BaseMultiVAE` to train

        train_dataset (MultimodalBaseDataset): The training dataset of type
            :class:`~multivae.data.datasets.MultimodalBaseDataset`

        eval_dataset (MultimodalBaseDataset): The evaluation dataset of type
            :class:`~multivae.data.datasets.MultimodalBaseDataset`

        training_config (BaseTrainerConfig): The training arguments summarizing the main
            parameters used for training. If None, a basic training instance of
            :class:`BaseTrainerConfig` is used. Default: None.

        callbacks (List[~pythae.trainers.training_callbacks.TrainingCallback]):
            A list of callbacks to use during training.
    """

    def __init__(
        self,
        model: JNFDcca,
        train_dataset: MultimodalBaseDataset,
        eval_dataset: Optional[MultimodalBaseDataset] = None,
        training_config: Optional[AddDccaTrainerConfig] = None,
        callbacks: List[TrainingCallback] = None,
    ):
        super().__init__(model, train_dataset, eval_dataset, training_config, callbacks)

        training_config.per_device_dcca_train_batch_size = min(
            len(train_dataset), training_config.per_device_dcca_train_batch_size
        )
        training_config.per_device_dcca_eval_batch_size = min(
            len(train_dataset), training_config.per_device_dcca_eval_batch_size
        )

        # TODO : maybe check that the chosen batch size is large enough and
        # that the chosen dcca batch size does'nt result in a large loss of data

        self.train_loader = self.get_train_dataloader_dcca(train_dataset)
        self.eval_loader = self.get_eval_dataloader_dcca(eval_dataset)
        self.training_config.learning_rate_vae = self.training_config.learning_rate
        self.training_config.learning_rate = self.training_config.learning_rate_dcca

    def get_train_dataloader_dcca(
        self, train_dataset: MultimodalBaseDataset
    ) -> torch.utils.data.DataLoader:
        if self.distributed:
            train_sampler = DistributedSampler(
                train_dataset, num_replicas=self.world_size, rank=self.rank
            )
        else:
            train_sampler = None
        return DataLoader(
            dataset=train_dataset,
            batch_size=self.training_config.per_device_dcca_train_batch_size,
            num_workers=self.training_config.train_dataloader_num_workers,
            shuffle=(train_sampler is None),
            sampler=train_sampler,
            drop_last=len(train_dataset)
            > self.training_config.per_device_dcca_train_batch_size,
        )

    def get_eval_dataloader_dcca(
        self, eval_dataset: MultimodalBaseDataset
    ) -> torch.utils.data.DataLoader:
        if self.distributed:
            eval_sampler = DistributedSampler(
                eval_dataset, num_replicas=self.world_size, rank=self.rank
            )
        else:
            eval_sampler = None
        return DataLoader(
            dataset=eval_dataset,
            batch_size=self.training_config.per_device_dcca_eval_batch_size,
            num_workers=self.training_config.eval_dataloader_num_workers,
            shuffle=(eval_sampler is None),
            sampler=eval_sampler,
            drop_last=len(eval_dataset)
            > self.training_config.per_device_dcca_eval_batch_size,
        )

    def prepare_train_step(self, epoch, best_train_loss, best_eval_loss):
        """
        Function to operate changes between train_steps such as resetting the optimizer and
        the best losses values.
        """

        if epoch == self.model.nb_epochs_dcca + 1:
            logger.info(
                "End the training of the DCCA module and move on to the joint VAE."
            )
            # Change the train and eval_loader and reset the optimizer

            self.train_loader = self.get_train_dataloader(self.train_dataset)
            self.eval_loader = self.get_eval_dataloader(self.eval_dataset)
            logger.info(
                f"Using train_loader with batch_size {len(self.train_dataset)// len(self.train_loader)} \n"
                + f"Using eval_loader with batch_size {len(self.eval_dataset)// len(self.eval_loader)} \n"
            )
            self.training_config.learning_rate = self.training_config.learning_rate_vae
            self.set_optimizer()
            self.set_scheduler()
            best_train_loss = 1e10
            best_eval_loss = 1e10

        elif epoch == self.model.nb_epochs_dcca + self.model.warmup + 1:
            # Just reset the optimizer
            logger.info(
                "End the training of the joint VAE and move on to learning the unimodal "
                " posteriors."
            )

            self.set_optimizer()
            self.set_scheduler()
            best_train_loss = 1e10
            best_eval_loss = 1e10

        return best_train_loss, best_eval_loss



from typing import Literal

from pydantic.dataclasses import dataclass

from ..base import BaseMultiVAEConfig


@dataclass
class MMVAEConfig(BaseMultiVAEConfig):

    """
    This class is the configuration class for the MMVAE model, from
    (Variational Mixture-of-Experts Autoencoders
    for Multi-Modal Deep Generative Models, Shi et al 2019,
    https://proceedings.neurips.cc/paper/2019/hash/0ae775a8cb3b499ad1fca944e6f5c836-Abstract.html)


    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. Default to True.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided,
            a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
        K (int) : the number of samples to use in the DreG loss. Default to 1.
        prior_and_posterior_dist (str) : The type of distribution to use for posterior and prior.
            Possible values ['laplace_with_softmax','normal'].
            Default to 'laplace_with_softmax' the posterior distribution that is used in
            the original paper.
        learn_prior (bool) : If True, the mean and variance of the prior are optimized during the
            training. Default to True.
        beta (float) : When using K = 1 (ElBO loss), the beta factor regularizes the divergence term.
            Default to 1.
    """

    K: int = 10
    prior_and_posterior_dist: Literal[
        "laplace_with_softmax", "normal"
    ] = "laplace_with_softmax"
    learn_prior: bool = True
    beta: float = 1.0



from itertools import combinations
from typing import List

import numpy as np
import torch
from pythae.models.base.base_utils import ModelOutput

from multivae.data import MultimodalBaseDataset
from multivae.data.utils import set_inputs_to_device
from multivae.samplers.base import BaseSampler

from ..base.evaluator_class import Evaluator
from .coherences_config import CoherenceEvaluatorConfig


class CoherenceEvaluator(Evaluator):
    """
    Class for computing coherences metrics.

    Args:

        model (BaseMultiVAE) : The model to evaluate.
        classifiers (dict) : A dictionary containing the pretrained classifiers to use for the coherence evaluation.
        test_dataset (MultimodalBaseDataset) : The dataset to use for computing the metrics.
        output (str) : The folder path to save metrics. The metrics will be saved in a metrics.txt file.
        eval_config (CoherencesEvaluatorConfig) : The configuration class to specify parameters for the evaluation.
        sampler (BaseSampler) : A custom sampler for computing the joint coherence. If None is provided, samples
            are generated from the prior.
    """

    def __init__(
        self,
        model,
        classifiers,
        test_dataset,
        output=None,
        eval_config=CoherenceEvaluatorConfig(),
        sampler: BaseSampler = None,
    ) -> None:
        super().__init__(model, test_dataset, output, eval_config, sampler)
        self.clfs = classifiers
        self.include_recon = eval_config.include_recon
        self.nb_samples_for_joint = eval_config.nb_samples_for_joint
        for k in self.clfs:
            self.clfs[k] = self.clfs[k].to(self.device).eval()

    def cross_coherences(self):
        """
        Computes all the coherences from one subset of modalities to another modality.

        Returns:
            float, float: The cross-coherences metric mean and std
        """

        modalities = list(self.model.encoders.keys())
        accs = []
        for n in range(1, self.model.n_modalities):
            subsets_of_size_n = combinations(
                modalities,
                n,
            )
            accs.append([])
            for s in subsets_of_size_n:
                s = list(s)
                subset_dict, mean_acc = self.all_accuracies_from_subset(s)
                self.metrics.update(subset_dict)
                accs[-1].append(mean_acc)
        mean_accs = [np.mean(l) for l in accs]
        std_accs = [np.std(l) for l in accs]

        for i in range(len(mean_accs)):
            self.logger.info(
                f"Conditional accuracies for {i+1} modalities : {mean_accs[i]} +- {std_accs[i]}"
            )
            self.metrics.update(
                {
                    f"mean_coherence_{i+1}": mean_accs[i],
                    f"std_coherence_{i+1}": std_accs[i],
                }
            )
        return mean_accs, std_accs

    def all_accuracies_from_subset(self, subset: List[str]):
        """
        Compute all the coherences generating from the modalities in subset to a modality
        that is not in subset.

        Args:
            subset (List[str]): The subset of modalities to consider.

        Returns:
            dict, float : The dictionary of all coherences from subset, and the mean coherence
        """
        accuracies = {}
        for batch in self.test_loader:

            if not hasattr(batch, "labels"):
                raise AttributeError(
                    "Cross-modal coherence can not be computed "
                    " on a dataset without labels"
                )
            elif batch.labels is None:
                raise AttributeError(
                    "Cross-modal coherence can not be computed "
                    " on a dataset without labels, but the provided dataset"
                    " has None instead of tensor labels"
                )

            batch = set_inputs_to_device(batch, device=self.device)
            pred_mods = [
                m
                for m in self.model.encoders
                if (m not in subset) or self.include_recon
            ]
            output = self.model.predict(batch, list(subset), pred_mods)
            for pred_m in pred_mods:
                preds = self.clfs[pred_m](output[pred_m])
                pred_labels = torch.argmax(preds, dim=1)
                try:
                    accuracies[f"subset_to_{pred_m}"] += torch.sum(
                        pred_labels.cpu() == batch.labels.cpu()
                    )
                except:
                    accuracies[f"subset_to_{pred_m}"] = torch.sum(
                        pred_labels.cpu() == batch.labels.cpu()
                    )

        acc = {k: accuracies[k].cpu().numpy() / self.n_data for k in accuracies}

        self.logger.info(f"Subset {subset} accuracies ")
        self.logger.info(acc.__str__())
        mean_pair_acc = np.mean(list(acc.values()))
        self.logger.info(f"Mean subset {subset} accuracies : " + str(mean_pair_acc))
        return acc, mean_pair_acc

    def joint_coherence(self):
        """
        Generate in all modalities from the prior and compute the percentage of samples where all modalities have the same
        labels.

        Returns:
            float: The joint coherence metric
        """

        all_labels = torch.tensor([]).to(self.device)
        samples_to_generate = self.nb_samples_for_joint

        # loop over batches
        while samples_to_generate > 0:
            batch_samples = min(self.batch_size, samples_to_generate)

            if self.sampler is None:
                output_prior = self.model.generate_from_prior(batch_samples)
            else:
                output_prior = self.sampler.sample(batch_samples)

            # set output to device
            output_prior.z = output_prior.z.to(self.device)
            if not output_prior.one_latent_space:
                for m in output_prior.modalities_z:
                    output_prior.modalities_z[m] = output_prior.modalities_z[m].to(
                        self.device
                    )

            # decode
            output_decode = self.model.decode(output_prior)
            labels = []
            for m in output_decode.keys():
                preds = self.clfs[m](output_decode[m])
                labels_m = torch.argmax(preds, dim=1)  # shape (nb_samples_for_joint,1)
                labels.append(labels_m)
            all_same_labels = torch.all(
                torch.stack([l == labels[0] for l in labels]), dim=0
            )
            all_labels = torch.cat((all_labels, all_same_labels.float()), dim=0)
            samples_to_generate -= batch_samples
        joint_coherence = all_labels.mean()

        sampler_name = "prior" if self.sampler is None else self.sampler.name
        self.logger.info(
            f"Joint coherence with sampler {sampler_name}: {joint_coherence}"
        )
        self.metrics.update({f"joint_coherence_{sampler_name}": joint_coherence})
        return joint_coherence

    def eval(self):
        self.cross_coherences()
        self.joint_coherence()

        self.log_to_wandb()

        return ModelOutput(**self.metrics)



from copy import deepcopy
from typing import List

import numpy as np
import torch
from pythae.models.base.base_model import BaseDecoder
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseEncoder
from pythae.models.nn.default_architectures import Encoder_VAE_MLP
from torch import nn

from multivae.models.base.base_config import BaseAEConfig


class Encoder_VAE_MLP_Style(BaseEncoder):
    """
    A basic MLP encoders with two output embeddings :

    returns :
        ModelOutput(embedding = ..,
                    style_embedding = ..,
                    log_covariance = ..,
                    style_log_covariance = ..,)
    """

    def __init__(self, args: dict):
        BaseEncoder.__init__(self)
        self.input_dim = args.input_dim
        self.latent_dim = args.latent_dim
        self.style_dim = args.style_dim

        layers = nn.ModuleList()

        layers.append(nn.Sequential(nn.Linear(np.prod(args.input_dim), 512), nn.ReLU()))

        self.layers = layers
        self.depth = len(layers)

        self.embedding = nn.Linear(512, self.latent_dim)
        self.log_var = nn.Linear(512, self.latent_dim)

        self.style_embedding = nn.Linear(512, self.style_dim)
        self.style_log_var = nn.Linear(512, self.style_dim)

    def forward(self, x, output_layer_levels: List[int] = None):
        output = ModelOutput()

        max_depth = self.depth

        if output_layer_levels is not None:
            assert all(
                self.depth >= levels > 0 or levels == -1
                for levels in output_layer_levels
            ), (
                f"Cannot output layer deeper than depth ({self.depth}). "
                f"Got ({output_layer_levels})."
            )

            if -1 in output_layer_levels:
                max_depth = self.depth
            else:
                max_depth = max(output_layer_levels)

        out = x.reshape(-1, np.prod(self.input_dim))

        for i in range(max_depth):
            out = self.layers[i](out)

            if output_layer_levels is not None:
                if i + 1 in output_layer_levels:
                    output[f"embedding_layer_{i+1}"] = out
            if i + 1 == self.depth:
                output["embedding"] = self.embedding(out)
                output["log_covariance"] = self.log_var(out)
                output["style_embedding"] = self.style_embedding(out)
                output["style_log_covariance"] = self.style_log_var(out)

        return output


def BaseDictEncoders(input_dims: dict, latent_dim: int):
    encoders = nn.ModuleDict()
    for mod in input_dims:
        config = BaseAEConfig(input_dim=input_dims[mod], latent_dim=latent_dim)
        encoders[mod] = Encoder_VAE_MLP(config)
    return encoders


def BaseDictEncoders_MultiLatents(
    input_dims: dict, latent_dim: int, modality_dims: dict
):
    encoders = nn.ModuleDict()
    for mod in input_dims:
        config = BaseAEConfig(
            input_dim=input_dims[mod],
            latent_dim=latent_dim,
            style_dim=modality_dims[mod],
        )
        encoders[mod] = Encoder_VAE_MLP_Style(config)
    return encoders


def BaseDictDecoders(input_dims: dict, latent_dim: int):
    decoders = nn.ModuleDict()
    for mod in input_dims:
        config = BaseAEConfig(input_dim=input_dims[mod], latent_dim=latent_dim)
        decoders[mod] = Decoder_AE_MLP(config)
    return decoders


def BaseDictDecodersMultiLatents(
    input_dims: dict, latent_dim: int, modality_dims: dict
):
    decoders = nn.ModuleDict()
    for mod in input_dims:
        config = BaseAEConfig(
            input_dim=input_dims[mod], latent_dim=latent_dim + modality_dims[mod]
        )
        decoders[mod] = Decoder_AE_MLP(config)
    return decoders


class Decoder_AE_MLP(BaseDecoder):
    # The same as in Pythae but allows for any input shape (*, latent_dim) with * containing any number of dimensions.
    def __init__(self, args: dict):
        BaseDecoder.__init__(self)

        self.input_dim = args.input_dim

        layers = nn.ModuleList()

        layers.append(nn.Sequential(nn.Linear(args.latent_dim, 512), nn.ReLU()))

        layers.append(
            nn.Sequential(nn.Linear(512, int(np.prod(args.input_dim))), nn.Sigmoid())
        )

        self.layers = layers
        self.depth = len(layers)

    def forward(self, z: torch.Tensor):
        output = ModelOutput()

        max_depth = self.depth
        out = z

        for i in range(max_depth):
            out = self.layers[i](out)
            if i + 1 == self.depth:
                output_shape = (*z.shape[:-1],) + self.input_dim
                output["reconstruction"] = out.reshape(output_shape)

        return output


class MultipleHeadJointEncoder(BaseEncoder):
    """
    A default instance of joint encoder created from copying the architectures for the unimodal encoders,
    concatenating their outputs and passing them through a unifying Multi-Layer-Perceptron.

        Args:
            dict_encoders (dict): Contains an instance of BaseEncoder for each modality (key).
            args (dict): config dictionary. Contains the latent dim.
            hidden_dim (int) : Default to 512.
            n_hidden_layers (int) : Default to 2.
    """

    def __init__(
        self,
        dict_encoders: nn.ModuleDict,
        args: dict,
        hidden_dim=512,
        n_hidden_layers=2,
        **kwargs,
    ):
        super().__init__()

        # Duplicate all the unimodal encoders with identical instances.
        self.encoders = nn.ModuleDict()
        self.joint_input_dim = 0
        for modality in dict_encoders:
            self.encoders[modality] = deepcopy(dict_encoders[modality])
            self.joint_input_dim += self.encoders[modality].latent_dim

        modules = [
            nn.Sequential(nn.Linear(self.joint_input_dim, hidden_dim), nn.ReLU(True))
        ]
        for i in range(n_hidden_layers - 1):
            modules.extend(
                [nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.ReLU(True))]
            )

        self.enc = nn.Sequential(*modules)
        self.fc1 = nn.Linear(hidden_dim, args.latent_dim)
        self.fc2 = nn.Linear(hidden_dim, args.latent_dim)

        self.latent_dim = args.latent_dim

    def forward(self, x: dict):
        """
        Implements the encoding of the data contained in x.

        Args:
            x (dict): Contains a tensor for each modality (key).
        """

        assert np.all(x.keys() == self.encoders.keys())

        modalities_outputs = []
        for mod in self.encoders:
            modalities_outputs.append(self.encoders[mod](x[mod])["embedding"])

        # Stack the modalities outputs
        concatened_outputs = torch.cat(modalities_outputs, dim=1)
        h = self.enc(concatened_outputs)
        embedding = self.fc1(h)
        log_covariance = self.fc2(h)
        output = ModelOutput(embedding=embedding, log_covariance=log_covariance)

        return output



"""

Sampler fitting an Masked Autoregressive Flow in the multimodal variational autoencoder's latent space.
This class is simply adapted from Pythae's iaf_sampler for the multimodal case. 


"""

from .maf_sampler import MAFSampler
from .maf_sampler_config import MAFSamplerConfig

__all__ = ["MAFSampler", "MAFSamplerConfig"]



import argparse
import os
from abc import ABC, abstractmethod
from itertools import chain, combinations

import torch
import torch.nn as nn
from torch.autograd import Variable


def poe(mu, logvar, eps=1e-8):
    var = torch.exp(logvar) + eps
    # precision of i-th Gaussian expert at point x
    T = 1.0 / var
    pd_mu = torch.sum(mu * T, dim=0) / torch.sum(T, dim=0)
    pd_var = 1.0 / torch.sum(T, dim=0)
    pd_logvar = torch.log(pd_var)
    return pd_mu, pd_logvar


def reweight_weights(w):
    w = w / w.sum()
    return w


def mixture_component_selection(flags, mus, logvars, w_modalities=None):
    # if not defined, take pre-defined weights
    num_components = mus.shape[0]
    num_samples = mus.shape[1]
    if w_modalities is None:
        w_modalities = torch.Tensor(flags.alpha_modalities).to(flags.device)
    idx_start = []
    idx_end = []
    for k in range(0, num_components):
        if k == 0:
            i_start = 0
        else:
            i_start = int(idx_end[k - 1])
        if k == w_modalities.shape[0] - 1:
            i_end = num_samples
        else:
            i_end = i_start + int(torch.floor(num_samples * w_modalities[k]))
        idx_start.append(i_start)
        idx_end.append(i_end)
    idx_end[-1] = num_samples
    mu_sel = torch.cat(
        [mus[k, idx_start[k] : idx_end[k], :] for k in range(w_modalities.shape[0])]
    )
    logvar_sel = torch.cat(
        [logvars[k, idx_start[k] : idx_end[k], :] for k in range(w_modalities.shape[0])]
    )
    return [mu_sel, logvar_sel]


def calc_kl_divergence(mu0, logvar0, mu1=None, logvar1=None, norm_value=None):
    if mu1 is None or logvar1 is None:
        KLD = -0.5 * torch.sum(1 - logvar0.exp() - mu0.pow(2) + logvar0)
    else:
        KLD = -0.5 * (
            torch.sum(
                1
                - logvar0.exp() / logvar1.exp()
                - (mu0 - mu1).pow(2) / logvar1.exp()
                + logvar0
                - logvar1
            )
        )
    if norm_value is not None:
        KLD = KLD / float(norm_value)
    return KLD


def calc_group_divergence_moe(flags, mus, logvars, weights, normalization=None):
    num_mods = mus.shape[0]
    num_samples = mus.shape[1]
    if normalization is not None:
        klds = torch.zeros(num_mods)
    else:
        klds = torch.zeros(num_mods, num_samples)
    klds = klds.to(flags.device)
    weights = weights.to(flags.device)
    for k in range(0, num_mods):
        kld_ind = calc_kl_divergence(
            mus[k, :, :], logvars[k, :, :], norm_value=normalization
        )
        if normalization is not None:
            klds[k] = kld_ind
        else:
            klds[k, :] = kld_ind
    if normalization is None:
        weights = weights.unsqueeze(1).repeat(1, num_samples)
    group_div = (weights * klds).sum(dim=0)
    return group_div, klds


class BaseMMVae(ABC, nn.Module):
    def __init__(self, flags, modalities):
        super(BaseMMVae, self).__init__()
        self.num_modalities = len(modalities.keys())
        self.flags = flags
        self.modalities = modalities
        self.subsets = self.set_subsets()
        self.set_fusion_functions()

        encoders = nn.ModuleDict()
        decoders = nn.ModuleDict()
        lhoods = dict()
        for m, m_key in enumerate(sorted(modalities.keys())):
            encoders[m_key] = modalities[m_key].encoder
            decoders[m_key] = modalities[m_key].decoder
            lhoods[m_key] = modalities[m_key].likelihood
        self.encoders = encoders
        self.decoders = decoders
        self.lhoods = lhoods

    def set_subsets(self):
        num_mods = len(list(self.modalities.keys()))

        """
        powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3)
        (1,2,3)
        """
        xs = list(self.modalities)
        # note we return an iterator rather than a list
        subsets_list = chain.from_iterable(
            combinations(xs, n) for n in range(len(xs) + 1)
        )
        subsets = dict()
        for k, mod_names in enumerate(subsets_list):
            mods = []
            for l, mod_name in enumerate(sorted(mod_names)):
                mods.append(self.modalities[mod_name])
            key = "_".join(sorted(mod_names))
            subsets[key] = mods
        return subsets

    def reparameterize(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        eps = Variable(std.data.new(std.size()).normal_())
        return eps.mul(std).add_(mu)

    def set_fusion_functions(self):
        self.modality_fusion = self.poe_fusion
        self.fusion_condition = self.fusion_condition_joint
        self.calc_joint_divergence = self.divergence_static_prior

    def divergence_static_prior(self, mus, logvars, weights=None):
        if weights is None:
            weights = self.weights
        weights = weights.clone()
        weights = reweight_weights(weights)
        div_measures = calc_group_divergence_moe(
            self.flags, mus, logvars, weights, normalization=self.flags.batch_size
        )
        divs = dict()
        divs["joint_divergence"] = div_measures[0]
        divs["individual_divs"] = div_measures[1]
        divs["dyn_prior"] = None
        return divs

    def moe_fusion(self, mus, logvars, weights=None):
        if weights is None:
            weights = self.weights
        weights = reweight_weights(weights)
        # mus = torch.cat(mus, dim=0);
        # logvars = torch.cat(logvars, dim=0);
        mu_moe, logvar_moe = mixture_component_selection(
            self.flags, mus, logvars, weights
        )
        return [mu_moe, logvar_moe]

    def poe_fusion(self, mus, logvars, weights=None):
        if self.flags.modality_poe or mus.shape[0] == len(self.modalities.keys()):
            num_samples = mus[0].shape[0]
            mus = torch.cat(
                (
                    mus,
                    torch.zeros(1, num_samples, self.flags.class_dim).to(
                        self.flags.device
                    ),
                ),
                dim=0,
            )
            logvars = torch.cat(
                (
                    logvars,
                    torch.zeros(1, num_samples, self.flags.class_dim).to(
                        self.flags.device
                    ),
                ),
                dim=0,
            )
        # mus = torch.cat(mus, dim=0);
        # logvars = torch.cat(logvars, dim=0);
        mu_poe, logvar_poe = poe(mus, logvars)
        return [mu_poe, logvar_poe]

    def fusion_condition_moe(self, subset, input_batch=None):
        if len(subset) == 1:
            return True
        else:
            return False

    def fusion_condition_poe(self, subset, input_batch=None):
        if len(subset) == len(input_batch.keys()):
            return True
        else:
            return False

    def fusion_condition_joint(self, subset, input_batch=None):
        return True

    def forward(self, input_batch):
        latents = self.inference(input_batch)
        results = dict()
        results["latents"] = latents
        results["group_distr"] = latents["joint"]
        class_embeddings = self.reparameterize(latents["joint"][0], latents["joint"][1])
        div = self.calc_joint_divergence(
            latents["mus"], latents["logvars"], latents["weights"]
        )
        for k, key in enumerate(div.keys()):
            results[key] = div[key]

        results_rec = dict()
        enc_mods = latents["modalities"]
        for m, m_key in enumerate(self.modalities.keys()):
            if m_key in input_batch.keys():
                m_s_mu, m_s_logvar = enc_mods[m_key + "_style"]
                if self.flags.factorized_representation:
                    m_s_embeddings = self.reparameterize(mu=m_s_mu, logvar=m_s_logvar)
                else:
                    m_s_embeddings = None
                m_rec = self.lhoods[m_key](
                    self.decoders[m_key](class_embeddings).reconstruction, 0.75
                )
                results_rec[m_key] = m_rec
        results["rec"] = results_rec
        return results

    def encode(self, input_batch):
        latents = dict()
        for m, m_key in enumerate(self.modalities.keys()):
            if m_key in input_batch.keys():
                i_m = input_batch[m_key]
                l = self.encoders[m_key](i_m)
                latents[m_key + "_style"] = [None, None]
                latents[m_key] = [l.embedding, l.log_covariance]
            else:
                latents[m_key + "_style"] = [None, None]
                latents[m_key] = [None, None]
        return latents

    def inference(self, input_batch, num_samples=None):
        if num_samples is None:
            num_samples = self.flags.batch_size
        latents = dict()
        enc_mods = self.encode(input_batch)
        latents["modalities"] = enc_mods
        mus = torch.Tensor().to(self.flags.device)
        logvars = torch.Tensor().to(self.flags.device)
        distr_subsets = dict()
        for k, s_key in enumerate(self.subsets.keys()):
            if s_key != "":
                mods = self.subsets[s_key]
                mus_subset = torch.Tensor().to(self.flags.device)
                logvars_subset = torch.Tensor().to(self.flags.device)
                mods_avail = True
                for m, mod in enumerate(mods):
                    if mod.name in input_batch.keys():
                        mus_subset = torch.cat(
                            (mus_subset, enc_mods[mod.name][0].unsqueeze(0)), dim=0
                        )
                        logvars_subset = torch.cat(
                            (logvars_subset, enc_mods[mod.name][1].unsqueeze(0)), dim=0
                        )
                    else:
                        mods_avail = False
                if mods_avail:
                    weights_subset = (1 / float(len(mus_subset))) * torch.ones(
                        len(mus_subset)
                    ).to(self.flags.device)
                    s_mu, s_logvar = self.modality_fusion(
                        mus_subset, logvars_subset, weights_subset
                    )
                    distr_subsets[s_key] = [s_mu, s_logvar]
                    if self.fusion_condition(mods, input_batch):
                        mus = torch.cat((mus, s_mu.unsqueeze(0)), dim=0)
                        logvars = torch.cat((logvars, s_logvar.unsqueeze(0)), dim=0)

        # weights = (1/float(len(mus)))*torch.ones(len(mus)).to(self.flags.device);
        weights = (1 / float(mus.shape[0])) * torch.ones(mus.shape[0]).to(
            self.flags.device
        )
        joint_mu, joint_logvar = self.moe_fusion(mus, logvars, weights)
        # mus = torch.cat(mus, dim=0);
        # logvars = torch.cat(logvars, dim=0);
        latents["mus"] = mus
        latents["logvars"] = logvars
        latents["weights"] = weights
        latents["joint"] = [joint_mu, joint_logvar]
        latents["subsets"] = distr_subsets
        return latents

    def generate(self, num_samples=None):
        if num_samples is None:
            num_samples = self.flags.batch_size

        mu = torch.zeros(num_samples, self.flags.class_dim).to(self.flags.device)
        logvar = torch.zeros(num_samples, self.flags.class_dim).to(self.flags.device)
        z_class = self.reparameterize(mu, logvar)
        z_styles = self.get_random_styles(num_samples)
        random_latents = {"content": z_class, "style": z_styles}
        random_samples = self.generate_from_latents(random_latents)
        return random_samples

    def generate_sufficient_statistics_from_latents(self, latents):
        suff_stats = dict()
        content = latents["content"]
        for m, m_key in enumerate(self.modalities.keys()):
            s = latents["style"][m_key]
            cg = self.lhoods[m_key](*self.decoders[m_key](s, content))
            suff_stats[m_key] = cg
        return suff_stats

    def generate_from_latents(self, latents):
        suff_stats = self.generate_sufficient_statistics_from_latents(latents)
        cond_gen = dict()
        for m, m_key in enumerate(latents["style"].keys()):
            cond_gen_m = suff_stats[m_key].mean
            cond_gen[m_key] = cond_gen_m
        return cond_gen

    def cond_generation(self, latent_distributions, num_samples=None):
        if num_samples is None:
            num_samples = self.flags.batch_size

        style_latents = self.get_random_styles(num_samples)
        cond_gen_samples = dict()
        for k, key in enumerate(latent_distributions.keys()):
            [mu, logvar] = latent_distributions[key]
            content_rep = self.reparameterize(mu=mu, logvar=logvar)
            latents = {"content": content_rep, "style": style_latents}
            cond_gen_samples[key] = self.generate_from_latents(latents)
        return cond_gen_samples

    def get_random_style_dists(self, num_samples):
        styles = dict()
        for k, m_key in enumerate(self.modalities.keys()):
            mod = self.modalities[m_key]
            s_mu = torch.zeros(num_samples, mod.style_dim).to(self.flags.device)
            s_logvar = torch.zeros(num_samples, mod.style_dim).to(self.flags.device)
            styles[m_key] = [s_mu, s_logvar]
        return styles

    def get_random_styles(self, num_samples):
        styles = dict()
        for k, m_key in enumerate(self.modalities.keys()):
            if self.flags.factorized_representation:
                mod = self.modalities[m_key]
                z_style = torch.randn(num_samples, mod.style_dim)
                z_style = z_style.to(self.flags.device)
            else:
                z_style = None
            styles[m_key] = z_style
        return styles


from multivae.models.nn.default_architectures import (
    BaseAEConfig,
    Decoder_AE_MLP,
    Encoder_VAE_MLP,
    ModelOutput,
)

flags = ModelOutput(
    factorized_representation=False,
    device="cpu",
    batch_size=2,
    class_dim=10,
    modality_poe=False,
)


modalities = dict(
    m0=ModelOutput(
        encoder=Encoder_VAE_MLP(BaseAEConfig(input_dim=(2,), latent_dim=10)),
        decoder=Decoder_AE_MLP(BaseAEConfig(input_dim=(2,), latent_dim=10)),
        likelihood=torch.distributions.Laplace,
        name="m0",
    ),
    m1=ModelOutput(
        encoder=Encoder_VAE_MLP(BaseAEConfig(input_dim=(3,), latent_dim=10)),
        decoder=Decoder_AE_MLP(BaseAEConfig(input_dim=(3,), latent_dim=10)),
        likelihood=torch.distributions.Laplace,
        name="m1",
    ),
)

model = BaseMMVae(flags, modalities)

data = dict(
    m0=torch.tensor([[1, 2], [3, 4]]).float(),
    m1=torch.tensor([[1, 2, 5], [3, 4, 6]]).float(),
)

output = model(data)

print(output)

print(output["rec"]["m0"].log_prob(data["m0"]).sum(-1).mean())
print(output["rec"]["m1"].log_prob(data["m1"]).sum(-1).mean())


from multivae.models import MoPoE, MoPoEConfig

config = MoPoEConfig(
    n_modalities=2,
    latent_dim=10,
    decoders_dist=dict(m0="laplace", m1="laplace"),
    decoder_dist_params=dict(m0=dict(scale=0.75), m1=dict(scale=0.75)),
)

encoders = dict(m0=modalities["m0"].encoder, m1=modalities["m1"].encoder)

decoders = dict(m0=modalities["m0"].decoder, m1=modalities["m1"].decoder)


model = MoPoE(config, encoders, decoders)
from multivae.data.datasets import MultimodalBaseDataset

print(model(MultimodalBaseDataset(data=data)))


def calc_log_prob(out_dist, target, norm_value):
    log_prob = out_dist.log_prob(target).sum()
    mean_val_logprob = log_prob / norm_value
    return mean_val_logprob


def calc_log_probs(flags, modalities, result, batch):
    mods = modalities
    log_probs = dict()
    weighted_log_prob = 0.0
    for m, m_key in enumerate(mods.keys()):
        mod = mods[m_key]
        log_probs[mod.name] = -calc_log_prob(
            result["rec"][mod.name], batch[mod.name], flags.batch_size
        )
        weighted_log_prob += log_probs[mod.name]
    return log_probs, weighted_log_prob


def calc_klds(flags, result):
    latents = result["latents"]["subsets"]
    klds = dict()
    for m, key in enumerate(latents.keys()):
        mu, logvar = latents[key]
        klds[key] = calc_kl_divergence(mu, logvar, norm_value=flags.batch_size)
    return klds


def basic_routine_epoch(flags, modalities, results, batch):
    log_probs, weighted_log_prob = calc_log_probs(flags, modalities, results, batch)
    group_divergence = results["joint_divergence"]

    klds = calc_klds(flags, results)

    kld_style = 0.0
    kld_content = group_divergence
    kld_weighted = 1 * kld_style + 1 * kld_content
    total_loss = 1 * weighted_log_prob + 1 * kld_weighted

    out_basic_routine = dict()
    out_basic_routine["weighted_log_probs"] = weighted_log_prob
    out_basic_routine["joint_divergence"] = kld_weighted

    out_basic_routine["total_loss"] = total_loss
    out_basic_routine["klds"] = klds
    return out_basic_routine


print(basic_routine_epoch(flags, modalities, output, data))



"""Implementation the TELBO algorithm from "Generative Models of Visually Grounded Imagination"
    (https://arxiv.org/abs/1705.10762)
"""

from .telbo_config import TELBOConfig
from .telbo_model import TELBO



import os

import numpy as np
import pytest
import torch
from pythae.models.base import BaseAEConfig
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.benchmarks.mnist.convnets import (
    Decoder_Conv_AE_MNIST,
    Encoder_Conv_AE_MNIST,
)
from torch import nn

from multivae.data.datasets import IncompleteDataset, MultimodalBaseDataset
from multivae.models import AutoConfig, AutoModel
from multivae.models.base import BaseMultiVAE, BaseMultiVAEConfig
from multivae.models.nn.default_architectures import Decoder_AE_MLP, Encoder_VAE_MLP
from multivae.trainers import BaseTrainerConfig

PATH = os.path.dirname(os.path.abspath(__file__))


class Test_BaseMultiVAE:
    @pytest.fixture
    def input_model1(self):
        model_config = BaseMultiVAEConfig(n_modalities=2, latent_dim=10)
        config = BaseAEConfig(input_dim=(10, 2), latent_dim=10)
        encoders = dict(
            mod1=Encoder_VAE_MLP(config), mod2=Encoder_Conv_AE_MNIST(config)
        )
        decoders = dict(mod1=Decoder_AE_MLP(config), mod2=Decoder_Conv_AE_MNIST(config))
        return dict(model_config=model_config, encoders=encoders, decoders=decoders)

    @pytest.fixture
    def input_model2(self):
        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=dict(mod1=(1, 2), mod2=(3, 4, 4))
        )

        return dict(model_config=model_config)

    def test1(self, input_model1):
        model = BaseMultiVAE(**input_model1)

        assert type(model.encoders) == nn.ModuleDict
        assert isinstance(model.encoders["mod1"], Encoder_VAE_MLP)
        assert isinstance(model.encoders["mod2"], Encoder_Conv_AE_MNIST)
        assert type(model.decoders) == nn.ModuleDict
        assert isinstance(model.decoders["mod1"], Decoder_AE_MLP)
        assert isinstance(model.decoders["mod2"], Decoder_Conv_AE_MNIST)
        assert model.latent_dim == input_model1["model_config"].latent_dim

    def test2(self, input_model2):
        model = BaseMultiVAE(**input_model2)

        assert type(model.encoders) == nn.ModuleDict
        assert isinstance(model.encoders["mod1"], Encoder_VAE_MLP)
        assert model.encoders["mod1"].input_dim == (1, 2)
        assert isinstance(model.encoders["mod2"], Encoder_VAE_MLP)
        assert model.encoders["mod2"].input_dim == (3, 4, 4)
        assert type(model.decoders) == nn.ModuleDict
        assert isinstance(model.decoders["mod1"], Decoder_AE_MLP)
        assert model.decoders["mod1"].input_dim == (1, 2)
        assert isinstance(model.decoders["mod2"], Decoder_AE_MLP)
        assert model.decoders["mod2"].input_dim == (3, 4, 4)
        assert model.latent_dim == input_model2["model_config"].latent_dim

    def test_raise_missing_input_dim(self, input_model1):
        # Test missing modality with
        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=dict(mod2=(3, 4, 4))
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config)

        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=None
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config)

        # Test missing modality with given encoders
        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=dict(mod2=(3, 4, 4))
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config, encoders=input_model1["encoders"])

        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=None
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config, encoders=input_model1["encoders"])

        model_config = BaseMultiVAEConfig(
            n_modalities=2,
            latent_dim=10,
            input_dims=None,
            uses_likelihood_rescaling=True,
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(
                model_config,
                encoders=input_model1["encoders"],
                decoders=input_model1["decoders"],
            )

    def test_raises_wrong_encoders(self, input_model1):
        # create dummy model
        model = BaseMultiVAE(**input_model1)

        encoders = dict(mod1=ModelOutput(), mod2=ModelOutput())

        with pytest.raises(AttributeError):
            model.set_encoders(encoders)

        decoders = dict(mod1=ModelOutput(), mod2=ModelOutput())

        with pytest.raises(AttributeError):
            model.set_decoders(decoders)

    def test_raises_key_error(self):
        model_config = BaseMultiVAEConfig(
            n_modalities=1, latent_dim=10, input_dims=dict(wrong_names=(3, 4, 4))
        )

        config = BaseAEConfig(input_dim=(10, 2), latent_dim=10)

        encoders = dict(mod1=Encoder_VAE_MLP(config))

        decoders = dict(mod1=Decoder_Conv_AE_MNIST(config))

        with pytest.raises(KeyError):
            BaseMultiVAE(model_config, encoders=encoders, decoders=decoders)

    def test_recon_dist(self):
        model_config = BaseMultiVAEConfig(
            n_modalities=3,
            latent_dim=10,
            input_dims=dict(mod1=(3,), mod2=(3, 4), mod3=(3, 4, 4)),
            decoders_dist=dict(mod1="normal", mod2="bernoulli", mod3="laplace"),
            decoder_dist_params=dict(
                mod1=dict(scale=12), mod2=None, mod3=dict(scale=31)
            ),
        )

        model = BaseMultiVAE(model_config)

        dumb_x1 = torch.randn(2, 3)
        dumb_x2 = torch.rand(2, 3, 4)
        dumb_x2_target = torch.randint(0, 2, (2, 3, 4)).float()
        dumb_x3 = torch.randn(2, 3, 4, 4)

        assert model.recon_log_probs["mod1"](dumb_x1, dumb_x1).shape == dumb_x1.shape
        assert (
            model.recon_log_probs["mod2"](dumb_x2, dumb_x2_target).shape
            == dumb_x2.shape
        )
        assert model.recon_log_probs["mod3"](dumb_x3, dumb_x3).shape == dumb_x3.shape

    def test_raises_sanity_check_flags(self):
        model_config = BaseMultiVAEConfig(
            n_modalities=2, latent_dim=10, input_dims=dict(mod1=(3, 4, 4), mod2=(3, 4))
        )

        config = BaseAEConfig(input_dim=(10, 2), latent_dim=10)

        encoders = dict(wrong_name=Encoder_VAE_MLP(config))

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config, encoders=encoders)

        decoders = dict(wrong_name=Decoder_Conv_AE_MNIST(config))

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config, decoders=decoders)

        encoders = dict(
            wrong_name1=Encoder_VAE_MLP(config), wrong_name2=Encoder_VAE_MLP(config)
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config, encoders=encoders)

        decoders = dict(
            wrong_name_1=Decoder_Conv_AE_MNIST(config),
            wrong_name_2=Decoder_Conv_AE_MNIST(config),
        )

        with pytest.raises(AttributeError):
            BaseMultiVAE(model_config, encoders=encoders, decoders=decoders)

    def test_raises_encode_error(self, input_model1):
        model = BaseMultiVAE(**input_model1)

        inputs = IncompleteDataset(
            data=dict(mod1=torch.randn(3, 10, 2), mod2=torch.rand(3, 10, 2)),
            masks=dict(mod1=torch.zeros((3,)), mod2=torch.ones((3,))),
        )

        with pytest.raises(AttributeError):
            model.encode(inputs, cond_mod="mod1")

    def test_decode_one_latent(self, input_model1):
        model = BaseMultiVAE(**input_model1)

        out = ModelOutput(
            z=torch.randn(3, input_model1["model_config"].latent_dim),
            one_latent_space=True,
        )
        output = model.decode(out, modalities=["mod1"])

        assert tuple(output.mod1.shape) == (3, 10, 2)

    def test_decode_several_latent(self):
        mod_latent = np.random.randint(1, 100)

        model_config = BaseMultiVAEConfig(n_modalities=2, latent_dim=10)
        config = BaseAEConfig(input_dim=(10, 2), latent_dim=10)
        encoders = dict(
            mod1=Encoder_VAE_MLP(config), mod2=Encoder_Conv_AE_MNIST(config)
        )
        config = BaseAEConfig(
            input_dim=(10, 2), latent_dim=mod_latent + config.latent_dim
        )

        decoders = dict(mod1=Decoder_AE_MLP(config), mod2=Decoder_Conv_AE_MNIST(config))

        model = BaseMultiVAE(model_config, encoders=encoders, decoders=decoders)

        out = ModelOutput(
            z=torch.randn(3, model_config.latent_dim),
            one_latent_space=False,
            modalities_z=dict(mod1=torch.randn(3, mod_latent)),
        )

        output = model.decode(out, modalities="mod1")

        assert tuple(output.mod1.shape) == (3, 10, 2)

    def test_raises_fwd_not_implemented(self, input_model1):
        model = BaseMultiVAE(**input_model1)

        with pytest.raises(NotImplementedError):
            model(None)

    def test_raises_nll_not_implemented(self, input_model1):
        model = BaseMultiVAE(**input_model1)

        with pytest.raises(NotImplementedError):
            model.compute_joint_nll(None)

    def test_generate_from_prior(self, input_model1):
        model = BaseMultiVAE(**input_model1)
        output = model.generate_from_prior(11)
        assert output.z.shape == (11, input_model1["model_config"].latent_dim)

    def test_dummy_model_saving(self, input_model1, tmpdir):
        model = BaseMultiVAE(**input_model1)

        assert not os.path.exists(os.path.join(tmpdir, "model_save"))
        model.save(os.path.join(tmpdir, "model_save"))
        assert os.path.exists(os.path.join(tmpdir, "model_save"))

        with pytest.raises(FileNotFoundError):
            model._load_model_config_from_folder(tmpdir)

        with pytest.raises(FileNotFoundError):
            model._load_model_weights_from_folder(tmpdir)

        torch.save({"wrong_key": torch.ones(2)}, os.path.join(tmpdir, "model.pt"))
        with pytest.raises(KeyError):
            model._load_model_weights_from_folder(tmpdir)

        with pytest.raises(FileNotFoundError):
            model._load_custom_archi_from_folder(tmpdir, "encoders")


class TestIntegrateAutoConfig:
    def test_autoconfig(self, tmpdir):
        model_config = BaseMultiVAEConfig(n_modalities=14, latent_dim=3)
        model_config.save_json(tmpdir, "model_config")
        reloaded_config = AutoConfig.from_json_file(
            os.path.join(tmpdir, "model_config.json")
        )

        assert model_config == reloaded_config

    def test_raises_not_handled(self, tmpdir):
        training_config = BaseTrainerConfig()
        training_config.save_json(tmpdir, "training_config")
        with pytest.raises(NameError):
            _ = AutoConfig.from_json_file(os.path.join(tmpdir, "training_config.json"))


class TestIntegrateAutoModel:
    def test_build_automodel(self):
        _ = AutoModel()

    def test_raises_not_handled(self):
        with pytest.raises(NameError):
            _ = AutoModel.load_from_folder(
                os.path.join(PATH, "tests_data", "wrong_model")
            )



from pydantic.dataclasses import dataclass

from ..base import BaseSamplerConfig


@dataclass
class IAFSamplerConfig(BaseSamplerConfig):
    """This is the IAF sampler model configuration instance.

    Parameters:
        n_made_blocks (int): The number of MADE model to consider in the IAF. Default: 2.
        n_hidden_in_made (int): The number of hidden layers in the MADE models. Default: 3.
        hidden_size (list): The number of unit in each hidder layer. The same number of units is
            used across the `n_hidden_in_made` and `n_made_blocks`. Default: 3.
        include_batch_norm (bool): Whether to include batch normalization after each
            :class:`~pythae.models.normalizing_flows.MADE` layers. Default: False.
    """

    n_made_blocks: int = 2
    n_hidden_in_made: int = 3
    hidden_size: int = 128
    include_batch_norm: bool = False



"""
**Abstract class**

This is the base Sampler architecture module from which all future samplers should inherit.
All samplers are adapted from the Pythae's samplers implementation for multimodal vaes.
"""

from .base_sampler import BaseSampler
from .base_sampler_config import BaseSamplerConfig

__all__ = ["BaseSampler", "BaseSamplerConfig"]



from itertools import combinations

import numpy as np
import torch
from pythae.models.base.base_utils import ModelOutput
from scipy import linalg
from torch.utils.data import DataLoader, TensorDataset
from torchvision.transforms import Resize

from multivae.data import MultimodalBaseDataset
from multivae.data.utils import set_inputs_to_device
from multivae.models.base import BaseMultiVAE
from multivae.samplers import BaseSampler

from ..base.evaluator_class import Evaluator
from .fids_config import FIDEvaluatorConfig
from .inception_networks import wrapper_inception

try:
    from tqdm import tqdm
except:
    tqdm = lambda x: x


class adapt_shape_for_fid(torch.nn.Module):
    """
    Transform an input so that each sample has three dimensions with three channels.
    (batch_size, 2,h,w). The input is assumed to be batched.
    """

    def __init__(self, resize=True, **kwargs) -> None:
        super().__init__(**kwargs)
        if resize:
            self.resize = Resize((299, 299))
        else:
            self.resize = None

    def forward(self, x):
        if len(x.shape) == 1:  # (n_data,)
            x = x.unsqueeze(1)
        if len(x.shape) == 2:  # (n_data, n)
            x = x.unsqueeze(1)
        if len(x.shape) == 3:  # (n_data, n, m)
            x = x.unsqueeze(1)
        if len(x.shape) == 4:
            if x.shape[1] == 1:
                # Add channels to have 3 channels
                x = torch.cat([x for _ in range(3)], dim=1)
            elif x.shape[1] == 2:
                n, ch, h, w = x.shape
                x = torch.cat([x, torch.zeros(n, 1, h, w)], dim=1)
            else:
                x = x[:, :3, :, :]

            if self.resize is not None:
                return self.resize(x)
            else:
                return x
        else:
            raise AttributeError("Can't visualize data with more than 3 dimensions")


class FIDEvaluator(Evaluator):
    """
    Class for computing likelihood metrics.

    Args:
        model (BaseMultiVAE) : The model to evaluate.
        classifiers (dict) : A dictionary containing the pretrained classifiers to use for the coherence evaluation.
        test_dataset (MultimodalBaseDataset) : The dataset to use for computing the metrics.
        output (str) : The folder path to save metrics. The metrics will be saved in a metrics.txt file.
        eval_config (EvaluatorConfig) : The configuration class to specify parameters for the evaluation.
        sampler (Basesampler) : The sampler used to generate from the latent space.
            If None is provided, the latent codes are generated from prior. Default to None.
        custom_encoder (torch.nn.Module) : If you desire, you can provide our own embedding architecture to use
            instead of the InceptionV3 model to compute Fréchet Distances.
            By default, the pretrained InceptionV3 network is used. Default to None.
        transform (torchvision.Transforms) : To apply to the images before computing the embeddings. If None is provided
            a default resizing to (3,299,299) is applied. Default to None.
    """

    def __init__(
        self,
        model: BaseMultiVAE,
        test_dataset,
        output=None,
        eval_config=FIDEvaluatorConfig(),
        sampler: BaseSampler = None,
        custom_encoder=None,
        transform=None,
    ) -> None:
        super().__init__(model, test_dataset, output, eval_config, sampler)

        if custom_encoder is not None:
            self.model_fd = custom_encoder
        else:
            self.model_fd = wrapper_inception(
                dims=eval_config.dims_inception,
                device=self.device,
                path_state_dict=eval_config.inception_weights_path,
            )
        if transform is not None:
            self.inception_transform = transform
        else:
            self.inception_transform = adapt_shape_for_fid()

    def get_frechet_distance(self, mod, generate_latent_function):
        """
        Calculates the activations of the pool_3 layer for all images.
        """
        self.model.eval()
        activations = [[], []]

        for batch in tqdm(self.test_loader):
            batch = set_inputs_to_device(batch, self.device)
            # Compute activations for true data
            data = self.inception_transform(batch.data[mod]).to(self.device)
            pred = self.model_fd(data)
            del data
            activations[0].append(pred)

            # Compute activations for generated data
            latents = generate_latent_function(len(pred), inputs=batch)
            latents.z = latents.z.to(self.device)

            samples = self.model.decode(latents, modalities=mod)
            data_gen = self.inception_transform(samples[mod])
            del samples
            pred_gen = self.model_fd(data_gen)
            activations[1].append(pred_gen)
            del data_gen

        activations = [np.concatenate(l, axis=0) for l in activations]

        # Compute activation statistics
        mus = [np.mean(act, axis=0) for act in activations]
        sigmas = [np.cov(act, rowvar=False) for act in activations]
        fd = self.calculate_frechet_distance(mus[0], sigmas[0], mus[1], sigmas[1])
        return fd

    def calculate_frechet_distance(self, mu1, sigma1, mu2, sigma2, eps=1e-6):
        r"""Numpy implementation of the Frechet Distance.
        The Frechet distance between two multivariate Gaussians :math:`X_1 \sim
        \mathcal{N}(\mu_1, C_1)`
        and :math:`X_2 \sim \mathcal{N}(\mu_2, C_2)` is
        :math:`d^2 = \lVert \mu_1 - \mu_2\rVert^2 + \mathrm{Tr}(C_1 + C_2 -
        2\sqrt{(C_1\cdot C_2)})`.
        Stable version by Dougal J. Sutherland.

        Args:
            mu1 (numpy.ndarray): Numpy array containing the activations of a layer of the
                    inception net (like returned by the function 'get_predictions')
                    for generated samples.
            mu2 (numpy.ndarray): The sample mean over activations, precalculated on an
                    representative data set.
            sigma1 (numpy.ndarray): The covariance matrix over activations for generated samples.
            sigma2 (numpy.ndarray): The covariance matrix over activations, precalculated on an
                    representative data set.

        Return:
            numpy.ndarray : The Frechet Distance.
        """
        mu1 = np.atleast_1d(mu1)
        mu2 = np.atleast_1d(mu2)

        sigma1 = np.atleast_2d(sigma1)
        sigma2 = np.atleast_2d(sigma2)

        assert mu1.shape == mu2.shape, (
            f"Training and test mean vectors have different lengths. mu1 has shape {mu1.shape}"
            f"whereas mu2 has shape {mu2.shape}"
        )
        assert (
            sigma1.shape == sigma2.shape
        ), "Training and test covariances have different dimensions"

        diff = mu1 - mu2

        # Product might be almost singular
        covmean, _ = linalg.sqrtm(sigma1.dot(sigma2), disp=False)
        if not np.isfinite(covmean).all():
            msg = (
                "fid calculation produces singular product; "
                "adding %s to diagonal of cov estimates"
            ) % eps
            self.logger.info(msg)
            offset = np.eye(sigma1.shape[0]) * eps
            covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

        # Numerical error might give slight imaginary component
        if np.iscomplexobj(covmean):
            if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
                m = np.max(np.abs(covmean.imag))
                raise ValueError("Imaginary component {}".format(m))
            covmean = covmean.real

        tr_covmean = np.trace(covmean)

        return diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean

    def unconditional_fids(self):
        """
        Generate data from prior or sampler fitted in the latent space
        and compute the FID for each modality.

        Returns:
            ~pythae.models.base.base_utils.ModelOutput: FIDs for all modalities.
        """
        output = dict()
        if self.sampler is None:
            generate_function = self.model.generate_from_prior
        else:
            generate_function = self.sampler.sample

        sampler_name = "prior" if self.sampler is None else self.sampler.name
        for mod in self.model.encoders:
            self.logger.info(f"Start computing FID for modality {mod}")
            fd = self.get_frechet_distance(mod, generate_function)
            output[f"fd_{mod}_sampler_{sampler_name}"] = fd
            self.logger.info(
                f"The FD for modality {mod} with sampler {sampler_name} is {fd}"
            )
        self.metrics.update(output)

        return ModelOutput(**output)

    def eval(self):
        self.unconditional_fids()
        self.log_to_wandb()

        return ModelOutput(**self.metrics)

    def compute_fid_from_conditional_generation(self, subset, gen_mod):
        """
        Generate samples from the conditional distribution conditioned on subset and compute
        Frechet distance for gen_mod.
        """

        generate_function = lambda n_samples, inputs: self.model.encode(
            inputs=inputs, cond_mod=subset
        )
        fd = self.get_frechet_distance(gen_mod, generate_function)
        self.logger.info(
            f"The FD for modality {gen_mod} computed from subset={subset} is {fd}"
        )

        self.metrics[f"Conditional FD from {subset} to {gen_mod}"] = fd
        return fd

    def compute_all_conditional_fids(self, gen_mod):
        """
        For all subsets in modalities \gen mod, compute the FID when generating
        images from the subsets.
        """

        modalities = [k for k in self.model.encoders if k != gen_mod]
        fds = []
        for n in range(1, len(modalities) + 1):
            s = modalities[:n]
            fd = self.compute_fid_from_conditional_generation(s, gen_mod)
            fds.append(fd)

        self.log_to_wandb()

        return ModelOutput(**self.metrics)



import math

import numpy as np
import torch
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseDecoder, BaseEncoder
from torch import nn


class PositionalEncoding(nn.Module):
    """Taken from torch/examples"""

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)
        )
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position * div_term)
        pe[0, :, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[seq_len, batch_size, embedding_dim]``
        """
        x = x + self.pe[:, : x.size(1)]
        return self.dropout(x)


class CubTextEncoder(BaseEncoder):
    def __init__(
        self,
        args,
        max_sentence_length: int,
        ntokens: int,
        embed_size: int = 512,
        nhead: int = 4,
        ff_size: int = 1024,
        n_layers: int = 4,
        dropout: float = 0.5,
    ):
        """
        A transformer-based encoder for text.

        Parameters:
            ntokens (int): Vocabulary size.
            embed_size (int): Size of the token embedding vectors. Default: 512
            nhead (int): Number of head in the MultiHeadAttention module. Default: 8
            ff_size (int): Number of units in the feedforward layers. Default: 1024
            n_layers (int): Number of Encoders layers in the TransformerEncoder. Default: 8
            dropout (float): Dropout rate. Default: 0.5
        """

        BaseEncoder.__init__(self)
        self.latent_dim = args.latent_dim

        self.embed_size = embed_size

        self.token_embedding = nn.Embedding(ntokens, embed_size)
        self.pos_encoder = PositionalEncoding(embed_size, dropout)
        encoder_layer = nn.TransformerEncoderLayer(
            embed_size, nhead, ff_size, dropout, batch_first=True
        )
        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, n_layers)

        self.mu = nn.Linear(embed_size * max_sentence_length, self.latent_dim)
        self.log_covariance = nn.Linear(
            embed_size * max_sentence_length, self.latent_dim
        )

        self.init_weights()

    def init_weights(self):
        initrange = 0.1
        nn.init.uniform_(self.token_embedding.weight, -initrange, initrange)

    def forward(self, inputs):
        src = inputs["tokens"]
        padding_mask = inputs["padding_mask"]

        src = self.token_embedding(src) * math.sqrt(self.embed_size)
        src = self.pos_encoder(src)
        transformer_output = self.transformer_encoder(
            src, src_key_padding_mask=padding_mask
        )
        output = ModelOutput(
            embedding=self.mu(transformer_output.reshape(src.shape[0], -1)),
            log_covariance=self.log_covariance(
                transformer_output.reshape(src.shape[0], -1)
            ),
            transformer_output=transformer_output,
        )
        return output


class CubTextDecoderMLP(BaseDecoder):
    def __init__(self, args: dict):
        """
        A simple MLP decoder for text.
        """
        BaseDecoder.__init__(self)

        self.input_dim = args.input_dim

        layers = nn.ModuleList()

        layers.append(nn.Sequential(nn.Linear(args.latent_dim, 512), nn.ReLU()))

        layers.append(nn.Sequential(nn.Linear(512, int(np.prod(args.input_dim)))))

        self.layers = layers
        self.depth = len(layers)

    def forward(self, z: torch.Tensor):
        output = ModelOutput()

        max_depth = self.depth
        out = z

        for i in range(max_depth):
            out = self.layers[i](out)
            if i + 1 == self.depth:
                output_shape = (*z.shape[:-1],) + self.input_dim
                output["reconstruction"] = out.reshape(output_shape)

        return output



"""
Implementation of the Multi-modal VAE model from "Multimodal Generative Models for Scalable
Weakly-Supervised Learning" (https://arxiv.org/abs/1802.05335).
"""

from .mvae_config import MVAEConfig
from .mvae_model import MVAE

__all__ = ["MVAE", "MVAEConfig"]



from pydantic.dataclasses import dataclass
from pythae.config import BaseConfig


@dataclass
class AutoConfig(BaseConfig):
    @classmethod
    def from_json_file(cls, json_path):
        """Creates a :class:`~multivae.config.BaseMultiVAEConfig` instance from a JSON config file. It
        builds automatically the correct config for any `pythae.models`.

        Args:
            json_path (str): The path to the json file containing all the parameters

        Returns:
            :class:`BaseMultiVAEConfig`: The created instance
        """

        config_dict = cls._dict_from_json(json_path)
        config_name = config_dict.pop("name")

        if config_name == "BaseMultiVAEConfig":
            from ..base import BaseMultiVAEConfig

            model_config = BaseMultiVAEConfig.from_json_file(json_path)

        elif config_name == "JMVAEConfig":
            from ..jmvae import JMVAEConfig

            model_config = JMVAEConfig.from_json_file(json_path)

        elif config_name == "JNFConfig":
            from ..jnf import JNFConfig

            model_config = JNFConfig.from_json_file(json_path)

        elif config_name == "MMVAEConfig":
            from ..mmvae import MMVAEConfig

            model_config = MMVAEConfig.from_json_file(json_path)
        elif config_name == "TELBOConfig":
            from ..telbo import TELBOConfig

            model_config = TELBOConfig.from_json_file(json_path)

        elif config_name == "MVAEConfig":
            from ..mvae import MVAEConfig

            model_config = MVAEConfig.from_json_file(json_path)

        elif config_name == "JNFDccaConfig":
            from ..jnf_dcca import JNFDccaConfig

            model_config = JNFDccaConfig.from_json_file(json_path)
        elif config_name == "MoPoEConfig":
            from ..mopoe import MoPoEConfig

            model_config = MoPoEConfig.from_json_file(json_path)

        elif config_name == "MVTCAEConfig":
            from ..mvtcae import MVTCAEConfig

            model_config = MVTCAEConfig.from_json_file(json_path)

        elif config_name == "MMVAEPlusConfig":
            from ..mmvaePlus import MMVAEPlusConfig

            model_config = MMVAEPlusConfig.from_json_file(json_path)

        else:
            raise NameError(
                "Cannot reload automatically the model configuration... "
                f"The model name in the `model_config.json may be corrupted. Got `{config_name}`"
            )

        return model_config



import logging
from typing import Union

import numpy as np
import torch
import torch.distributions as dist
import torch.nn as nn
import torch.nn.functional as F
from pythae.models.base.base_utils import ModelOutput
from torch.distributions import Laplace, Normal

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.models.nn.default_architectures import (
    BaseDictDecodersMultiLatents,
    BaseDictEncoders_MultiLatents,
)

from ..base import BaseMultiVAE
from .mmvaePlus_config import MMVAEPlusConfig

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class MMVAEPlus(BaseMultiVAE):

    """
    The MMVAE+ model.

    Args:
        model_config (MMVAEPlusConfig): An instance of MMVAEConfig in which any model's
            parameters is made available.

        encoders (Dict[str, ~pythae.models.nn.base_architectures.BaseEncoder]): A dictionary containing
            the modalities names and the encoders for each modality. Each encoder is an instance of
            Pythae's BaseEncoder. Default: None.

        decoders (Dict[str, ~pythae.models.nn.base_architectures.BaseDecoder]): A dictionary containing
            the modalities names and the decoders for each modality. Each decoder is an instance of
            Pythae's BaseDecoder.
    """

    def __init__(
        self,
        model_config: MMVAEPlusConfig,
        encoders: dict = None,
        decoders: dict = None,
    ):
        if model_config.modalities_specific_dim is None:
            raise AttributeError(
                "The modalities_specific_dim attribute must"
                " be provided in the model config."
            )

        super().__init__(model_config, encoders, decoders)

        self.K = model_config.K
        if model_config.prior_and_posterior_dist == "laplace_with_softmax":
            self.post_dist = Laplace
            self.prior_dist = Laplace
        elif model_config.prior_and_posterior_dist == "normal":
            self.post_dist = Normal
            self.prior_dist = Normal
        else:
            raise AttributeError(
                " The posterior_dist parameter must be "
                " either 'laplace_with_softmax' or 'normal'. "
                f" {model_config.prior_and_posterior_dist} was provided."
            )

        # Set the priors for shared and private spaces.
        self.mean_priors = torch.nn.ParameterDict()
        self.logvars_priors = torch.nn.ParameterDict()
        self.beta = model_config.beta
        self.modalities_specific_dim = model_config.modalities_specific_dim
        self.reconstruction_option = model_config.reconstruction_option
        self.multiple_latent_spaces = True
        self.style_dims = {m: self.modalities_specific_dim for m in self.encoders}

        # Add the private and shared latents priors.

        # modality specific priors (referred to as r distributions in paper)
        for mod in list(self.encoders.keys()):
            self.mean_priors[mod] = torch.nn.Parameter(
                torch.zeros(1, model_config.modalities_specific_dim),
                requires_grad=False,
            )
            self.logvars_priors[mod] = torch.nn.Parameter(
                torch.zeros(1, model_config.modalities_specific_dim),
                requires_grad=model_config.learn_modality_prior,
            )

        # general prior (for the entire latent code) referred to as p in the paper
        self.mean_priors["shared"] = torch.nn.Parameter(
            torch.zeros(
                1, model_config.latent_dim + model_config.modalities_specific_dim
            ),
            requires_grad=False,
        )
        self.logvars_priors["shared"] = torch.nn.Parameter(
            torch.zeros(
                1, model_config.latent_dim + model_config.modalities_specific_dim
            ),
            requires_grad=model_config.learn_shared_prior,
        )

        self.model_name = "MMVAEPlus"

    def log_var_to_std(self, log_var):
        """
        For latent distributions parameters, transform the log covariance to the
        standard deviation of the distribution either applying softmax or not.
        This follows the original implementation.
        """

        if self.model_config.prior_and_posterior_dist == "laplace_with_softmax":
            return F.softmax(log_var, dim=-1) * log_var.size(-1) + 1e-6
        else:
            return torch.exp(0.5 * log_var)

    def forward(self, inputs: MultimodalBaseDataset, **kwargs):
        # TODO : maybe implement a minibatch strategy for stashing the gradients before
        # backpropagation when using a large number k.
        # Also, I've only implemented the dreg_looser loss but it may be nice to offer other options.

        # First compute all the encodings for all modalities
        embeddings = {}
        qu_xs = {}
        qw_xs = {}
        qu_xs_detach = {}
        qw_xs_detach = {}

        reconstructions = {}

        compute_loss = kwargs.pop("compute_loss", True)
        detailed_output = kwargs.pop("detailed_output", False)
        K = kwargs.pop("K", self.K)

        for cond_mod in self.encoders:
            output = self.encoders[cond_mod](inputs.data[cond_mod])
            mu, log_var = output.embedding, output.log_covariance
            mu_style = output.style_embedding
            log_var_style = output.style_log_covariance

            sigma = self.log_var_to_std(log_var)
            sigma_style = self.log_var_to_std(log_var_style)

            # Shared latent variable
            qu_x = self.post_dist(mu, sigma)
            u_x = qu_x.rsample([K])

            # Private latent variable
            qw_x = self.post_dist(mu_style, sigma_style)
            w_x = qw_x.rsample([K])

            # The DREG loss uses detached parameters in the loss computation afterwards.
            qu_x_detach = self.post_dist(mu.detach(), sigma.detach())
            qw_x_detach = self.post_dist(mu_style.detach(), sigma_style.detach())

            # Then compute all the cross-modal reconstructions
            reconstructions[cond_mod] = {}

            for recon_mod in self.decoders:
                # Self-reconstruction
                if recon_mod == cond_mod:
                    z_x = torch.cat([u_x, w_x], dim=-1)

                # Cross modal reconstruction
                else:
                    # only keep the shared latent and generate private from prior

                    mu_prior_mod = torch.cat(
                        [self.mean_priors[recon_mod]] * len(mu), axis=0
                    )
                    sigma_prior_mod = torch.cat(
                        [self.log_var_to_std(self.logvars_priors[recon_mod])] * len(mu),
                        axis=0,
                    )

                    w = self.prior_dist(
                        mu_prior_mod,
                        sigma_prior_mod,
                    ).rsample([K])
                    # print(w.shape)
                    # print(u_x.shape,w.shape)
                    z_x = torch.cat([u_x, w], dim=-1)
                # Decode
                # print(z_x.shape)
                decoder = self.decoders[recon_mod]
                recon = decoder(z_x)["reconstruction"]

                reconstructions[cond_mod][recon_mod] = recon

            qu_xs[cond_mod] = qu_x
            qu_xs_detach[cond_mod] = qu_x_detach
            qw_xs[cond_mod] = qw_x
            qw_xs_detach[cond_mod] = qw_x_detach
            embeddings[cond_mod] = dict(u=u_x, w=w_x)

        # Compute DREG loss
        if compute_loss:
            # loss_output = self.dreg_looser(
            #     qu_xs_detach, qw_xs_detach, embeddings, reconstructions, inputs
            # )
            loss_output = self.dreg_looser(
                qu_xs, qw_xs, embeddings, reconstructions, inputs
            )

        else:
            loss_output = ModelOutput()
        if detailed_output:
            loss_output["zss"] = embeddings
            loss_output["recon"] = reconstructions

        return loss_output

    @property
    def pz_params(self):
        """From the prior mean and log_covariance, return the mean and standard
        deviation, either applying softmax or not depending on the choice of prior
        distribution.

        Returns:
            tuple: mean, std
        """
        mean = self.mean_priors["shared"]
        if self.model_config.prior_and_posterior_dist == "laplace_with_softmax":
            std = F.softmax(
                self.logvars_priors["shared"], dim=-1
            ) * self.logvars_priors["shared"].size(-1)
        else:
            std = torch.exp(0.5 * self.logvars_priors["shared"])
        return mean, std

    def dreg_looser(self, qu_xs, qw_xs, embeddings, reconstructions, inputs):
        """
        The objective function used in the original implementation.
        it is said to use the Dreg estimator but I am not sure it is done correctly since it doesn't
        detach the parameters of the posteriors.
        """

        if hasattr(inputs, "masks"):
            # Compute the number of available modalities per sample
            n_mods_sample = torch.sum(
                torch.stack(tuple(inputs.masks.values())).int(), dim=0
            )
        else:
            n_mods_sample = torch.tensor([self.n_modalities])

        lws = []
        zss = []
        for mod in embeddings:
            # Log-probabilities to the prior are the same as in MMVAE
            u = embeddings[mod]["u"]  # (K, n_batch, latent_dim)
            w = embeddings[mod]["w"]  # (K, n_batch, latent_dim)
            n_mods_sample = n_mods_sample.to(u.device)
            prior = self.prior_dist(*self.pz_params)
            z = torch.cat([u, w], dim=-1)
            lpz = prior.log_prob(z).sum(-1)

            # For the shared latent variable it is the same
            if hasattr(inputs, "masks"):
                lqu_x = torch.stack(
                    [
                        qu_xs[m].log_prob(u).sum(-1) * inputs.masks[m].float()
                        for m in qu_xs
                    ]
                )  # n_modalities,K,nbatch
            else:
                lqu_x = torch.stack(
                    [qu_xs[m].log_prob(u).sum(-1) for m in qu_xs]
                )  # n_modalities,K,nbatch

            lqu_x = torch.logsumexp(lqu_x, dim=0) - torch.log(
                n_mods_sample
            )  # log_mean_exp

            # Then we have to add the modality specific posterior
            lqw_x = qw_xs[mod].log_prob(w).sum(-1)

            # The reconstructions are the same
            lpx_z = 0
            for recon_mod in reconstructions[mod]:
                x_recon = reconstructions[mod][recon_mod]
                K, n_batch = x_recon.shape[0], x_recon.shape[1]
                lpx_z_mod = (
                    self.recon_log_probs[recon_mod](x_recon, inputs.data[recon_mod])
                    .view(K, n_batch, -1)
                    .mul(self.rescale_factors[recon_mod])
                    .sum(-1)
                )

                if hasattr(inputs, "masks"):
                    # cancel unavailable modalities
                    lpx_z_mod *= inputs.masks[recon_mod].float()

                lpx_z += lpx_z_mod

            lw = lpx_z + self.beta * (lpz - lqu_x - lqw_x)

            if hasattr(inputs, "masks"):
                # cancel unavailable modalities
                lw *= inputs.masks[mod].float()

            lws.append(lw)
            zss.append(z)

        lws = torch.stack(lws)  # (n_modalities, K, n_batch)
        zss = torch.stack(zss)  # (n_modalities, K, n_batch,latent_dim)
        with torch.no_grad():
            grad_wt = (lws - torch.logsumexp(lws, 1, keepdim=True)).exp()
            if zss.requires_grad:  # True except when we are in eval mode
                zss.register_hook(lambda grad: grad_wt.unsqueeze(-1) * grad)

        lws = (grad_wt * lws).sum(0) / n_mods_sample  # mean over modalities

        return ModelOutput(loss=-lws.sum(), metrics=dict(mean_loss_batch=-lws.mean()))

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ):
        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod
        if all([s in self.encoders.keys() for s in cond_mod]):
            # For the conditioning modalities we compute all the embeddings
            encoders_outputs = {m: self.encoders[m](inputs.data[m]) for m in cond_mod}

            # Choose one of the conditioning modalities at random to sample the shared information.
            random_mod = np.random.choice(cond_mod)

            # Sample the shared latent code
            mu = encoders_outputs[random_mod].embedding
            log_var = encoders_outputs[random_mod].log_covariance
            sigma = self.log_var_to_std(log_var)

            qz_x = self.post_dist(mu, sigma)
            sample_shape = torch.Size([]) if N == 1 else torch.Size([N])
            z = qz_x.rsample(sample_shape)

            flatten = kwargs.pop("flatten", False)
            if flatten:
                z = z.reshape(-1, self.latent_dim)

            # Modality specific encodings : given by encoders for conditioning modalities
            # Sampling from the priors for the rest of the modalities.

            style_z = dict()

            for m in self.encoders:
                if m not in cond_mod:
                    # Sample from priors parameters.
                    if self.reconstruction_option == "single_prior":
                        mu_m = self.mean_priors[m]
                        logvar_m = self.logvars_priors[m]

                    if self.reconstruction_option == "joint_prior":
                        mu_m = self.mean_priors["shared"][:, self.latent_dim :]
                        logvar_m = self.logvars_priors["shared"][:, self.latent_dim :]

                    mu_m = torch.cat([mu_m] * len(mu), dim=0)
                    logvar_m = torch.cat([logvar_m] * len(mu), dim=0)

                else:
                    # Sample from posteriors parameters
                    mu_m = encoders_outputs[m].style_embedding
                    logvar_m = encoders_outputs[m].style_log_covariance

                sigma_m = self.log_var_to_std(logvar_m)
                style_z[m] = self.post_dist(mu_m, sigma_m).rsample(sample_shape)
                if flatten:
                    style_z[m] = style_z[m].reshape(-1, self.modalities_specific_dim)

            return ModelOutput(z=z, one_latent_space=False, modalities_z=style_z)

    def compute_joint_nll(
        self, inputs: MultimodalBaseDataset, K: int = 1000, batch_size_K: int = 100
    ):
        """
        Return the estimated negative log-likelihood summed over the inputs.
        The negative log-likelihood is estimated using importance sampling.

        Args:
            inputs : the data to compute the joint likelihood

        """
        raise (NotImplementedError)

        # First compute all the parameters of the joint posterior q(z|x,y)
        post_params = []
        for cond_mod in self.encoders:
            output = self.encoders[cond_mod](inputs.data[cond_mod])
            mu, log_var = output.embedding, output.log_covariance
            sigma = self.log_var_to_std(log_var)
            post_params.append((mu, sigma))

        z_joint = self.encode(inputs, N=K).z
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, nb_channels, w, h)
                    x_m = inputs.data[mod][i]  # (nb_channels, w, h)

                    lpx_zs += (
                        self.recon_log_probs[mod](recon, x_m)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = self.prior_dist(*self.pz_params)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xs = [self.post_dist(p[0][i], p[1][i]) for p in post_params]
                lqz_xy = torch.logsumexp(
                    torch.stack([q.log_prob(latents).sum(-1) for q in qz_xs]), dim=0
                ) - np.log(self.n_modalities)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll

    def generate_from_prior(self, n_samples, **kwargs):
        sample_shape = [n_samples] if n_samples > 1 else []
        z = self.prior_dist(*self.pz_params).rsample(sample_shape).to(self.device)
        return ModelOutput(z=z.squeeze(), one_latent_space=True)

    def default_encoders(self, model_config) -> nn.ModuleDict:
        return BaseDictEncoders_MultiLatents(
            input_dims=model_config.input_dims,
            latent_dim=model_config.latent_dim,
            modality_dims={
                m: model_config.modalities_specific_dim
                for m in self.model_config.input_dims
            },
        )

    def default_decoders(self, model_config) -> nn.ModuleDict:
        return BaseDictDecodersMultiLatents(
            input_dims=model_config.input_dims,
            latent_dim=model_config.latent_dim,
            modality_dims={
                m: model_config.modalities_specific_dim for m in model_config.input_dims
            },
        )



from pydantic.dataclasses import dataclass

from ..base.evaluator_config import EvaluatorConfig


@dataclass
class FIDEvaluatorConfig(EvaluatorConfig):
    """

    Config class for the evaluation of the coherences module.

    Args:
        batch_size (int) : The batch size to use in the evaluation. Default to 512
        wandb_path (str) : The user can provide the path of the wandb run with a
            format 'entity/projet_name/run_id' where the metrics should be logged.
            For an existing run (the training run), the info can be found in the training dir (in wandb_info.json)
            at the end of training (if wandb was used) or on the hugging_face webpage of the run.
            Otherwise the user can create a new wandb run and get the path with :
                `
                import wandb
                run = wandb.init(entity = your_entity, project=your_project)
                wandb_path = run.path

            If None are provided, the metrics are not logged on wandb.
            Default to None.
        inception_weights_path (str) : The path to InceptionV3 weights. Default to
            '../fid_model/model.pt'.
        dims_inception (int) : Select the embedding layer of the Inception network
            defined by its output_size. Default to 2048.

    """

    inception_weights_path: str = "../fid_model/model.pt"
    dims_inception: int = 2048



"""
Implementation of the Joint Multimodal VAE model from the paper "Joint Multimodal Learning with Deep
Generative Models" (http://arxiv.org/abs/1611.01891).
"""

from .jmvae_config import JMVAEConfig
from .jmvae_model import JMVAE

__all__ = ["JMVAEConfig", "JMVAE"]



import numpy as np
import torch
from pythae.models.base.base_model import BaseDecoder, BaseEncoder
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.benchmarks.utils import ResBlock
from torch import nn

from multivae.models.base.base_config import BaseAEConfig


class Flatten(torch.nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)


class Unflatten(torch.nn.Module):
    def __init__(self, ndims):
        super(Unflatten, self).__init__()
        self.ndims = ndims

    def forward(self, x):
        return x.view(x.size(0), *self.ndims)


class EncoderConvMMNIST(BaseEncoder):
    """
    Adopted from:
    https://www.cs.toronto.edu/~lczhang/360/lec/w05/autoencoder.html
    """

    def __init__(self, model_config: BaseAEConfig):
        super(EncoderConvMMNIST, self).__init__()
        self.latent_dim = model_config.latent_dim
        self.shared_encoder = nn.Sequential(  # input shape (3, 28, 28)
            nn.Conv2d(
                3, 32, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (32, 14, 14)
            nn.ReLU(),
            nn.Conv2d(
                32, 64, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (64, 7, 7)
            nn.ReLU(),
            nn.Conv2d(
                64, 128, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (128, 4, 4)
            nn.ReLU(),
            # nn.Conv2d(128, 256, kernel_size=3, stride=2, padding=1, bias=True),  # -> (256, 2, 2)
            # nn.ReLU(),
            nn.Flatten(),
            nn.Linear(2048, self.latent_dim),  # -> (ndim_private + ndim_shared)
            nn.ReLU(),
        )

        # content branch
        self.class_mu = nn.Linear(self.latent_dim, self.latent_dim)
        self.class_logvar = nn.Linear(self.latent_dim, self.latent_dim)

    def forward(self, x):
        h = self.shared_encoder(x)
        return ModelOutput(
            embedding=self.class_mu(h), log_covariance=self.class_logvar(h)
        )


class EncoderConvMMNIST_adapted(BaseEncoder):
    """
    Adapt so that it works with DCCA and models with multiple latent spaces.
    """

    def __init__(self, model_config: BaseAEConfig):
        super(EncoderConvMMNIST_adapted, self).__init__()
        self.latent_dim = model_config.latent_dim
        self.style_dim = model_config.style_dim
        self.shared_encoder = nn.Sequential(  # input shape (3, 28, 28)
            nn.Conv2d(
                3, 32, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (32, 14, 14)
            nn.ReLU(),
            nn.Conv2d(
                32, 64, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (64, 7, 7)
            nn.ReLU(),
            nn.Conv2d(
                64, 128, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (128, 4, 4)
            nn.ReLU(),
        )

        # content branch
        self.class_mu = nn.Conv2d(128, self.latent_dim, 4, 2, 0)
        self.class_logvar = nn.Conv2d(128, self.latent_dim, 4, 2, 0)

    def forward(self, x):
        h = self.shared_encoder(x)
        return ModelOutput(
            embedding=self.class_mu(h).squeeze(),
            log_covariance=self.class_logvar(h).squeeze(),
        )


class EncoderConvMMNIST_multilatents(BaseEncoder):
    """
    Adapt so that it works with multiple latent spaces models.
    """

    def __init__(self, model_config: BaseAEConfig):
        super(EncoderConvMMNIST_multilatents, self).__init__()
        self.latent_dim = model_config.latent_dim
        self.style_dim = model_config.style_dim
        self.encoder_class = nn.Sequential(  # input shape (3, 28, 28)
            nn.Conv2d(
                3, 32, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (32, 14, 14)
            nn.ReLU(),
            nn.Conv2d(
                32, 64, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (64, 7, 7)
            nn.ReLU(),
            nn.Conv2d(
                64, 128, kernel_size=3, stride=2, padding=1, bias=True
            ),  # -> (128, 4, 4)
            nn.ReLU(),
        )

        # content branch
        self.class_mu = nn.Conv2d(128, self.latent_dim, 4, 2, 0)
        self.class_logvar = nn.Conv2d(128, self.latent_dim, 4, 2, 0)

        if self.style_dim > 0:
            self.encoder_style = nn.Sequential(  # input shape (3, 28, 28)
                nn.Conv2d(
                    3, 32, kernel_size=3, stride=2, padding=1, bias=True
                ),  # -> (32, 14, 14)
                nn.ReLU(),
                nn.Conv2d(
                    32, 64, kernel_size=3, stride=2, padding=1, bias=True
                ),  # -> (64, 7, 7)
                nn.ReLU(),
                nn.Conv2d(
                    64, 128, kernel_size=3, stride=2, padding=1, bias=True
                ),  # -> (128, 4, 4)
                nn.ReLU(),
            )

            self.style_mu = nn.Conv2d(128, self.style_dim, 4, 2, 0)
            self.style_logvar = nn.Conv2d(128, self.style_dim, 4, 2, 0)

    def forward(self, x):
        output = ModelOutput()
        # content branch
        h_class = self.encoder_class(x)
        output["embedding"] = self.class_mu(h_class).squeeze()
        output["log_covariance"] = self.class_logvar(h_class).squeeze()

        if self.style_dim > 0:
            # style branch
            h_style = self.encoder_style(x)
            output["style_embedding"] = self.style_mu(h_style).squeeze()
            output["style_log_covariance"] = self.style_logvar(h_style).squeeze()

        return output


class DecoderConvMMNIST(BaseDecoder):
    """
    Adopted from:
    https://www.cs.toronto.edu/~lczhang/360/lec/w05/autoencoder.html
    """

    def __init__(self, model_config: BaseAEConfig):
        super(DecoderConvMMNIST, self).__init__()
        self.latent_dim = model_config.latent_dim
        self.decoder = nn.Sequential(
            nn.Linear(self.latent_dim, 2048),  # -> (2048)
            nn.ReLU(),
            Unflatten((128, 4, 4)),  # -> (128, 4, 4)
            nn.ConvTranspose2d(
                128,
                64,
                kernel_size=3,
                stride=2,
                padding=1,
            ),  # -> (128, 4, 4)
            nn.ReLU(),
            nn.ConvTranspose2d(
                64, 32, kernel_size=3, stride=2, padding=1, output_padding=1
            ),  # -> (32, 14, 14)
            nn.ReLU(),
            nn.ConvTranspose2d(
                32, 3, kernel_size=3, stride=2, padding=1, output_padding=1
            ),  # -> (3, 28, 28)
        )

    def forward(self, z):
        x_hat = self.decoder(z.view(-1, z.size(-1)))
        # x_hat = torch.sigmoid(x_hat)
        x_hat = x_hat.view(*z.size()[:-1], *x_hat.size()[1:])
        return ModelOutput(
            reconstruction=x_hat
        )  # NOTE: consider learning scale param, too


class Encoder_ResNet_VAE_MMNIST(BaseEncoder):
    """
    A ResNet encoder suited for MNIST and Variational Autoencoder-based models.

    It can be built as follows:

    .. code-block::

        >>> from pythae.models.nn.benchmarks.mnist import Encoder_ResNet_VAE_MNIST
        >>> from pythae.models import VAEConfig
        >>> model_config = VAEConfig(input_dim=(1, 28, 28), latent_dim=16)
        >>> encoder = Encoder_ResNet_VAE_MNIST(model_config)
        >>> encoder
        ... Encoder_ResNet_VAE_MNIST(
        ...   (layers): ModuleList(
        ...     (0): Sequential(
        ...       (0): Conv2d(1, 64, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1))
        ...     )
        ...     (1): Sequential(
        ...       (0): Conv2d(64, 128, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1))
        ...     )
        ...     (2): Sequential(
        ...       (0): Conv2d(128, 128, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        ...     )
        ...     (3): Sequential(
        ...       (0): ResBlock(
        ...         (conv_block): Sequential(
        ...           (0): ReLU()
        ...           (1): Conv2d(128, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        ...           (2): ReLU()
        ...           (3): Conv2d(32, 128, kernel_size=(1, 1), stride=(1, 1))
        ...         )
        ...       )
        ...       (1): ResBlock(
        ...         (conv_block): Sequential(
        ...           (0): ReLU()
        ...           (1): Conv2d(128, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        ...           (2): ReLU()
        ...           (3): Conv2d(32, 128, kernel_size=(1, 1), stride=(1, 1))
        ...         )
        ...       )
        ...     )
        ...   )
        ...   (embedding): Linear(in_features=2048, out_features=16, bias=True)
        ...   (log_var): Linear(in_features=2048, out_features=16, bias=True)
        ... )


    and then passed to a :class:`pythae.models` instance

        >>> from pythae.models import VAE
        >>> model = VAE(model_config=model_config, encoder=encoder)
        >>> model.encoder == encoder
        ... True

    .. note::

        Please note that this encoder is only suitable for Autoencoder based models since it only
        outputs the embeddings of the input data under the key `embedding`.

        .. code-block::

            >>> import torch
            >>> input = torch.rand(2, 1, 28, 28)
            >>> out = encoder(input)
            >>> out.embedding.shape
            ... torch.Size([2, 16])

    """

    def __init__(self, args: BaseAEConfig):
        BaseEncoder.__init__(self)

        self.input_dim = (3, 28, 28)
        self.latent_dim = args.latent_dim
        self.n_channels = 3

        layers = nn.ModuleList()

        layers.append(nn.Sequential(nn.Conv2d(self.n_channels, 64, 4, 2, padding=1)))

        layers.append(nn.Sequential(nn.Conv2d(64, 128, 4, 2, padding=1)))

        layers.append(nn.Sequential(nn.Conv2d(128, 128, 3, 2, padding=1)))

        layers.append(
            nn.Sequential(
                ResBlock(in_channels=128, out_channels=32),
                ResBlock(in_channels=128, out_channels=32),
            )
        )

        self.layers = layers
        self.depth = len(layers)

        self.embedding = nn.Linear(128 * 4 * 4, args.latent_dim)
        self.log_var = nn.Linear(128 * 4 * 4, args.latent_dim)

    def forward(self, x: torch.Tensor):
        """Forward method

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            ModelOutput: An instance of ModelOutput containing the embeddings of the input data
            under the key `embedding`. Optional: The outputs of the layers specified in
            `output_layer_levels` arguments are available under the keys `embedding_layer_i` where
            i is the layer's level."""
        output = ModelOutput()

        max_depth = self.depth

        out = x

        for i in range(max_depth):
            out = self.layers[i](out)
            if i + 1 == self.depth:
                output["embedding"] = self.embedding(out.reshape(x.shape[0], -1))
                output["log_covariance"] = self.log_var(out.reshape(x.shape[0], -1))

        return output


class Decoder_ResNet_AE_MMNIST(BaseDecoder):
    """
    A ResNet decoder suited for MNIST and Autoencoder-based
    models.

    .. code-block::

        >>> from pythae.models.nn.benchmarks.mnist import Decoder_ResNet_AE_MNIST
        >>> from pythae.models import VAEConfig
        >>> model_config = VAEConfig(input_dim=(1, 28, 28), latent_dim=16)
        >>> decoder = Decoder_ResNet_AE_MNIST(model_config)
        >>> decoder
        ... Decoder_ResNet_AE_MNIST(
        ...   (layers): ModuleList(
        ...     (0): Linear(in_features=16, out_features=2048, bias=True)
        ...     (1): ConvTranspose2d(128, 128, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        ...     (2): Sequential(
        ...       (0): ResBlock(
        ...         (conv_block): Sequential(
        ...           (0): ReLU()
        ...           (1): Conv2d(128, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        ...           (2): ReLU()
        ...           (3): Conv2d(32, 128, kernel_size=(1, 1), stride=(1, 1))
        ...         )
        ...       )
        ...       (1): ResBlock(
        ...         (conv_block): Sequential(
        ...           (0): ReLU()
        ...           (1): Conv2d(128, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        ...           (2): ReLU()
        ...           (3): Conv2d(32, 128, kernel_size=(1, 1), stride=(1, 1))
        ...         )
        ...       )
        ...       (2): ReLU()
        ...     )
        ...     (3): Sequential(
        ...       (0): ConvTranspose2d(128, 64, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), output_padding=(1, 1))
        ...       (1): ReLU()
        ...     )
        ...     (4): Sequential(
        ...       (0): ConvTranspose2d(64, 1, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), output_padding=(1, 1))
        ...       (1): Sigmoid()
        ...     )
        ...   )
        ... )



    and then passed to a :class:`pythae.models` instance

        >>> from pythae.models import VAE
        >>> model = VAE(model_config=model_config, decoder=decoder)
        >>> model.decoder == decoder
        ... True

    .. note::

        Please note that this decoder is suitable for **all** models.

        .. code-block::

            >>> import torch
            >>> input = torch.randn(2, 16)
            >>> out = decoder(input)
            >>> out.reconstruction.shape
            ... torch.Size([2, 1, 28, 28])
    """

    def __init__(self, args: BaseAEConfig):
        BaseDecoder.__init__(self)

        self.input_dim = (3, 28, 28)
        self.latent_dim = args.latent_dim
        self.n_channels = 3

        layers = nn.ModuleList()

        layers.append(nn.Linear(args.latent_dim, 128 * 4 * 4))

        layers.append(nn.ConvTranspose2d(128, 128, 3, 2, padding=1))

        layers.append(
            nn.Sequential(
                ResBlock(in_channels=128, out_channels=32),
                ResBlock(in_channels=128, out_channels=32),
                nn.ReLU(),
            )
        )

        layers.append(
            nn.Sequential(
                nn.ConvTranspose2d(128, 64, 3, 2, padding=1, output_padding=1),
                nn.ReLU(),
            )
        )

        layers.append(
            nn.Sequential(
                nn.ConvTranspose2d(
                    64, self.n_channels, 3, 2, padding=1, output_padding=1
                ),
                nn.Sigmoid(),
            )
        )

        self.layers = layers
        self.depth = len(layers)

    def forward(self, z: torch.Tensor):
        """Forward method

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            ModelOutput: An instance of ModelOutput containing the reconstruction of the latent code
            under the key `reconstruction`. Optional: The outputs of the layers specified in
            `output_layer_levels` arguments are available under the keys `reconstruction_layer_i`
            where i is the layer's level.
        """
        output = ModelOutput()

        max_depth = self.depth

        out = z

        for i in range(max_depth):
            out = self.layers[i](out)

            if i == 0:
                output_shape = (np.prod(z.shape[:-1]),) + (128, 4, 4)
                out = out.reshape(*output_shape)

            if i + 1 == self.depth:
                output_shape = (*z.shape[:-1],) + (3, 28, 28)
                output["reconstruction"] = out.reshape(output_shape)

        return output



from typing import Union

from numpy import ndarray
from pythae.data.datasets import Dataset, DatasetOutput
from torch import Tensor


class MultimodalBaseDataset(Dataset):
    """This class is the Base class for datasets. A ``__getitem__`` is redefined and outputs a
    python dictionnary with the keys corresponding to `data` and `labels`. This Class should be
    used for any new data sets.

    Args:
        data (dict) : A dictionary containing the modalities' name and a tensor or numpy array for each modality.
        labels (Union[torch.Tensor, numpy.ndarray]) : A torch.Tensor or numpy.ndarray instance containing the labels.
    """

    def __init__(self, data: dict, labels: Union[Tensor, ndarray] = None):
        self.labels = labels
        self.data = data

    def __len__(self):
        length = len(self.data[list(self.data)[0]])
        for m in self.data:
            if len(self.data[m]) != length:
                raise AttributeError(
                    "The size of the provided datasets doesn't correspond between modalities!"
                )

        return length

    def __getitem__(self, index):
        """Generates one sample of data

        Args:
            index (int): The index of the data in the Dataset

        Returns:
            (dict): A dictionnary with the keys 'data' and 'labels' and corresponding
            torch.Tensor
        """
        # Select sample
        X = {modality: self.data[modality][index] for modality in self.data}
        y = self.labels[index] if self.labels is not None else None

        return DatasetOutput(data=X, labels=y)


class IncompleteDataset(MultimodalBaseDataset):
    """This class is the Base class for datasets with incomplete data.
    We add a field masks to indicate which data samples are available.
    This is used with models compatible with weakly supervised learning such as
    the MVAE. A ``__getitem__`` is redefined and outputs a python dictionnary with the keys
    corresponding to `data` and `labels`. This Class should be used for any new data sets.

    Args:
        data (dict[str, torch.Tensor]) : A dictionary containing the modalities' name and a tensor or numpy array for each modality.
        masks (dict[str, torch.Tensor]) : A dictionary containing the modalities'name and a boolean tensor of the same lenght
            as the data tensor in the data dictionary. For each modality, the mask tensor indicates if a sample
            is available. The unavailable samples are assumed to have been filled by random/or zeros values in
            the data dictionary.
        labels (Union[Tensor, numpy.ndarray]) : A torch.Tensor or numpy.ndarray instance containing the labels.

    """

    def __init__(self, data: dict, masks: dict, labels: Tensor = None) -> None:
        
        self.data = data
        self.masks = masks
        self.labels = labels

    def __len__(self):
        length = len(self.data[list(self.data)[0]])
        for m in self.data:
            if len(self.data[m]) != length or len(self.masks[m]) != length:
                raise AttributeError(
                    "The size of the provided datasets/masks doesn't correspond between modalities!"
                )

        return length

    def __getitem__(self, index):
        """Generates one sample of data

        Args:
            index (int): The index of the data in the Dataset

        Returns:
            (dict): A dictionnary with the keys 'data' and 'labels' and corresponding
            torch.Tensor
        """
        # Select sample
        X = {modality: self.data[modality][index] for modality in self.data}
        y = self.labels[index] if self.labels is not None else None
        m = {modality: self.masks[modality][index] for modality in self.masks}

        return DatasetOutput(data=X, labels=y, masks=m)



import os

import numpy as np
import pytest
from pythae.data.datasets import DatasetOutput

from multivae.data.datasets.base import MultimodalBaseDataset


class Test:
    @pytest.fixture
    def input_dataset_test(self):
        data = dict(
            mod1=np.array([[1, 2], [4, 5]]),
            mod2=np.array([[67, 2, 3], [1, 2, 3]]),
        )
        labels = np.array([0, 1])
        return dict(data=data, labels=labels)

    def test_create_dataset(self, input_dataset_test):
        dataset = MultimodalBaseDataset(**input_dataset_test)

        sample = dataset[0]
        assert type(sample) == DatasetOutput
        assert np.all(sample["data"]["mod1"] == np.array([1, 2]))
        assert np.all(sample["data"]["mod2"] == np.array([67, 2, 3]))
        assert sample["labels"] == 0



"""Contains relevant metrics classes that can be used with any model."""

from .coherences.coherences import CoherenceEvaluator
from .coherences.coherences_config import CoherenceEvaluatorConfig
from .fids.fids import FIDEvaluator
from .fids.fids_config import FIDEvaluatorConfig
from .latent_clustering.clustering_class import Clustering, ClusteringConfig
from .likelihoods.likelihoods import LikelihoodsEvaluator, LikelihoodsEvaluatorConfig
from .visualization import Visualization, VisualizationConfig

__all__ = [
    "CoherenceEvaluator",
    "CoherenceEvaluatorConfig",
    "FIDEvaluator",
    "FIDEvaluatorConfig",
    "LikelihoodsEvaluator",
    "LikelihoodsEvaluatorConfig",
    "Visualization",
    "VisualizationConfig",
    "Clustering",
    "ClusteringConfig",
]



from pydantic.dataclasses import dataclass

from ..base import BaseMultiVAEConfig


@dataclass
class MVAEConfig(BaseMultiVAEConfig):

    """
    Config class for the MVAE model from 'Multimodal Generative Models for Scalable Weakly-Supervised Learning'.
    https://proceedings.neurips.cc/paper/2018/hash/1102a326d5f7c9e04fc3c89d0ede88c9-Abstract.html

    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. Default to False.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided,
            a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
        use_subsampling (bool) : If True, we use the subsampling paradigm described in the article, not only taking the
            joint ELBO but also the unimodal ELBOs and k random subset elbos. This is useful when training with
            a complete dataset but should be set to False when the dataset is already incomplete.
            Default to True.
        k (int) : The number of subsets to use in the objective. The MVAE objective is the sum
            of the unimodal ELBOs, the joint ELBO and of k random subset ELBOs. Default to 0.
        warmup (int) : If warmup > 0, the MVAE model uses annealing during the first warmup epochs.
            In the objective, the KL terms are weighted by a factor beta that is linearly brought
            to 1 during the first warmup epochs. Default to 10.
        beta (float) : The scaling factor for the divergence term. Default to 1.


    """

    use_subsampling: bool = True
    k: int = 0
    warmup: int = 10
    beta: float = 1
